/*
**	Command & Conquer Renegade(tm)
**	Copyright 2025 Electronic Arts Inc.
**
**	This program is free software: you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation, either version 3 of the License, or
**	(at your option) any later version.
**
**	This program is distributed in the hope that it will be useful,
**	but WITHOUT ANY WARRANTY; without even the implied warranty of
**	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**	GNU General Public License for more details.
**
**	You should have received a copy of the GNU General Public License
**	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "missingtexture.h"
#include "texture.h"
#include "dx8wrapper.h"
#include <D3dx8core.h>

static unsigned missing_image_width=128;
static unsigned missing_image_height=128;
static unsigned missing_image_depth=24;
static unsigned missing_image_run_count=7331;

extern unsigned char missing_image_run_lengths[];
extern unsigned missing_image_color_values[];

static IDirect3DTexture8 * _MissingTexture = NULL;

IDirect3DTexture8* MissingTexture::_Get_Missing_Texture()
{
	WWASSERT(_MissingTexture);
	_MissingTexture->AddRef();
	return _MissingTexture;
}

IDirect3DSurface8* MissingTexture::_Create_Missing_Surface()
{
	IDirect3DSurface8 *texture_surface = NULL;
	DX8_ErrorCode(_MissingTexture->GetSurfaceLevel(0, &texture_surface));
	D3DSURFACE_DESC texture_surface_desc;
	::ZeroMemory(&texture_surface_desc, sizeof(D3DSURFACE_DESC));
	DX8_ErrorCode(texture_surface->GetDesc(&texture_surface_desc));
	
	IDirect3DSurface8 *surface = NULL;	
	DX8CALL(CreateImageSurface(
		texture_surface_desc.Width, 
		texture_surface_desc.Height, 
		texture_surface_desc.Format, 
		&surface));
	DX8CALL(CopyRects(texture_surface, NULL, 0, surface, NULL));
	texture_surface->Release();
	return surface;
}

void MissingTexture::_Init()
{
	WWASSERT(!_MissingTexture);

	IDirect3DTexture8* tex=DX8Wrapper::_Create_DX8_Texture(
		missing_image_width,
		missing_image_height,
		WW3D_FORMAT_A8R8G8B8,
		TextureClass::MIP_LEVELS_ALL);

	D3DLOCKED_RECT locked_rect;
	RECT rect;
	rect.left=0;
	rect.right=missing_image_width;
	rect.top=0;
	rect.bottom=missing_image_height;
	DX8_ErrorCode(
		tex->LockRect(
			0,
			&locked_rect,
			&rect,
			0));

	unsigned x=0;
	unsigned y=missing_image_height;
	unsigned *buffer=(unsigned*)locked_rect.pBits;
	unsigned pixcount=0;
	for (unsigned run=0;run<missing_image_run_count;++run) {
		unsigned color=missing_image_color_values[run];
		int run_length=missing_image_run_lengths[run]+1;
		for (int i=0;i<run_length;++i) {
			*buffer++=color;
			pixcount++;
			x++;
			if (x>=missing_image_width) {
				y--;
				x=0;
				buffer=(unsigned*)locked_rect.pBits;
				buffer+=locked_rect.Pitch/sizeof(unsigned)*y;
			}
		}
	}

	DX8_ErrorCode(tex->UnlockRect(0));

	for (unsigned i=1;i<tex->GetLevelCount();++i) {
		IDirect3DSurface8 *src,*dst;
		DX8_ErrorCode(tex->GetSurfaceLevel(i-1,&src));
		DX8_ErrorCode(tex->GetSurfaceLevel(i,&dst));

		DX8_ErrorCode(D3DXLoadSurfaceFromSurface(
			dst,
			NULL,	// palette
			NULL,	// rect
			src,
			NULL,	// palette
			NULL,	// rect
			D3DX_FILTER_BOX,	// box is good for 2:1 filtering
			0));

		src->Release();
		dst->Release();
	}

	_MissingTexture=tex;
}

void MissingTexture::_Deinit()
{
	_MissingTexture->Release();
	_MissingTexture=0;
}

unsigned char missing_image_run_lengths[]={
0x7f,0x7f,0x7f,0x7f,0x04,0x00,0x03,0x00,
0x00,0x73,0x04,0x00,0x04,0x00,0x00,0x72,
0x04,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x72,0x04,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x3c,0x00,0x00,0x33,0x04,
0x00,0x03,0x00,0x00,0x00,0x3c,0x00,0x00,
0x33,0x04,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x3d,0x00,0x00,0x00,0x32,0x04,0x00,
0x00,0x00,0x00,0x00,0x01,0x00,0x3c,0x00,
0x00,0x00,0x00,0x31,0x04,0x00,0x04,0x00,
0x20,0x00,0x00,0x00,0x19,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x2c,0x04,
0x00,0x03,0x00,0x00,0x20,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x16,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x2b,0x2c,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x11,0x00,0x00,0x00,
0x00,0x00,0x01,0x00,0x00,0x2b,0x2f,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x0d,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x01,0x2a,0x32,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x01,0x00,0x00,0x00,
0x00,0x00,0x00,0x03,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x01,0x2a,0x07,0x00,
0x00,0x2a,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x29,0x06,0x00,0x01,0x00,
0x27,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x29,0x05,0x00,0x00,
0x01,0x00,0x00,0x25,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x29,0x05,0x00,0x00,0x01,0x00,0x00,0x22,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x29,
0x04,0x00,0x00,0x03,0x00,0x00,0x1f,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x27,0x03,0x00,0x00,0x00,0x00,
0x04,0x1d,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x26,0x07,0x01,0x1f,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x24,0x07,0x01,
0x1d,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x23,0x07,0x01,
0x1c,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x21,0x07,0x01,0x1b,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x01,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x20,0x07,0x01,0x19,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x1f,0x07,0x01,0x18,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x1e,0x07,
0x01,0x17,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x1d,0x07,0x01,0x16,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x01,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x1c,0x07,
0x01,0x15,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x01,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x01,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x1b,0x07,0x01,0x14,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x01,
0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x1a,0x07,0x01,0x13,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x01,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x01,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x19,0x07,0x01,
0x13,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x18,0x07,0x01,0x12,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x17,0x07,0x01,0x11,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x01,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x17,0x07,0x01,0x11,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x16,0x07,0x01,0x10,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x16,0x07,0x01,0x10,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x15,0x07,
0x01,0x0f,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x14,0x07,0x01,0x0e,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x01,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x14,0x07,0x01,0x0e,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x02,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x01,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x13,0x07,0x01,0x0d,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x03,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x02,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x13,
0x07,0x01,0x0b,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x03,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x01,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x12,0x07,
0x01,0x08,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x03,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x02,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x12,
0x07,0x01,0x08,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x03,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x02,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x12,0x07,0x01,0x08,0x00,0x01,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x03,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x02,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x12,0x07,0x01,0x0a,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x03,0x02,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x02,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x11,0x07,
0x01,0x07,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x01,0x00,0x00,
0x01,0x00,0x01,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x03,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x11,0x07,0x01,0x08,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x03,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x01,0x00,0x00,0x00,
0x00,0x01,0x00,0x02,0x00,0x00,0x00,0x01,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x11,
0x07,0x01,0x07,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x03,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x02,0x00,0x01,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x10,0x07,0x01,0x06,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x01,0x01,0x00,0x00,0x00,0x01,0x01,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x01,
0x00,0x00,0x00,0x02,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x10,0x07,0x01,0x05,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x01,0x00,0x00,
0x00,0x00,0x00,0x00,0x01,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x01,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x10,0x07,0x01,0x05,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x01,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x01,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x01,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x10,0x07,0x01,0x0a,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x01,0x00,0x01,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x02,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x01,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x0f,0x07,0x01,
0x0a,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x02,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x01,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x01,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x0f,0x07,0x01,0x0a,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x02,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x01,0x01,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x01,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x0f,0x07,0x01,0x0a,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x01,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x02,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x01,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x01,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x0f,0x07,0x01,0x0a,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x01,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x01,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x01,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x0f,
0x07,0x01,0x0a,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x01,0x00,0x00,0x00,0x01,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x02,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x0f,0x07,0x01,0x0a,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x01,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x01,
0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x01,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x0f,0x07,0x01,
0x08,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x01,0x01,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x01,0x02,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x0f,0x07,0x01,0x08,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x01,0x01,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x01,0x00,0x00,0x00,0x00,0x01,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x0f,0x07,
0x01,0x0a,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x02,0x01,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x00,
0x01,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x01,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x0f,
0x07,0x01,0x0a,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x01,0x01,0x00,0x00,
0x00,0x00,0x00,0x01,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x01,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x02,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x0f,0x07,0x01,0x0a,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x01,0x01,0x00,0x00,0x00,
0x00,0x00,0x00,0x02,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x01,0x00,0x00,0x00,0x00,0x00,0x01,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x01,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x0d,0x07,0x01,0x08,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x01,0x00,0x00,0x02,0x01,0x00,
0x00,0x00,0x00,0x00,0x02,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x01,0x00,0x01,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x02,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x0d,0x07,0x01,0x09,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x02,
0x00,0x00,0x00,0x00,0x00,0x01,0x00,0x01,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x01,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x02,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x01,0x01,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x0d,0x07,0x01,0x0a,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x04,0x01,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x01,
0x01,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x0d,0x07,0x01,
0x07,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x02,0x01,0x00,0x00,0x00,0x00,0x00,
0x01,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x01,0x00,0x01,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x0a,0x07,
0x01,0x06,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x04,0x00,0x00,0x00,0x00,0x00,
0x02,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x01,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x01,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x0b,0x07,0x01,
0x07,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x01,0x00,0x00,0x04,
0x00,0x00,0x00,0x00,0x00,0x02,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x01,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x0b,0x07,0x01,0x0a,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x03,0x00,0x00,0x00,0x00,
0x00,0x00,0x02,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x01,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x0d,0x07,
0x01,0x0a,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x04,0x00,0x00,
0x00,0x00,0x00,0x03,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x01,0x01,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x0f,0x07,0x01,0x0c,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x03,0x01,0x00,0x00,0x00,0x00,0x03,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x01,
0x01,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x11,0x07,
0x01,0x0c,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x03,0x01,0x00,0x00,0x00,0x00,
0x05,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x01,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x11,
0x07,0x01,0x0c,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x01,0x03,0x00,0x00,0x00,
0x00,0x05,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x01,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x12,
0x07,0x01,0x0d,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x05,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x01,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x12,0x07,0x01,0x0d,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x07,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x13,0x07,0x01,0x0e,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x07,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x13,0x07,0x01,0x0e,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x06,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x13,0x07,0x01,0x0f,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x06,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x01,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x14,0x07,0x01,0x0f,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x05,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x01,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x01,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x14,
0x07,0x01,0x0f,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x04,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x01,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x15,0x07,
0x01,0x10,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x01,0x03,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x01,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x16,0x07,0x01,0x11,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x02,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x01,0x00,0x00,0x00,0x01,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x16,0x07,0x01,0x11,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x02,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x17,0x07,
0x01,0x12,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x02,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x01,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x18,0x07,0x01,0x13,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x01,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x18,0x07,0x01,
0x14,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x19,
0x07,0x01,0x15,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x1a,0x07,0x01,0x15,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x02,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x02,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x1b,0x07,0x01,0x16,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x01,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x02,0x01,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x1c,0x07,
0x01,0x17,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x05,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x1d,0x07,0x01,
0x18,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x05,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x1e,0x04,0x04,0x00,0x00,0x00,
0x00,0x15,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x01,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x1f,0x04,0x00,
0x00,0x03,0x00,0x00,0x18,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x20,0x05,
0x00,0x00,0x01,0x00,0x00,0x1a,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x21,0x05,0x00,
0x00,0x01,0x00,0x00,0x1c,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x23,0x06,0x00,0x00,0x00,0x00,0x1e,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x24,0x07,0x00,0x00,0x20,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x01,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x26,0x2c,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x28,0x2e,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x2a,0x2e,
0x01,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x2c,0x06,0x00,0x01,0x00,
0x23,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x2f,0x06,0x00,0x01,0x00,0x23,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x2f,0x06,0x00,0x01,0x00,0x23,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x03,0x00,0x00,0x00,0x01,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x2e,0x06,0x00,
0x01,0x00,0x24,0x01,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x0b,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x2b,0x06,0x00,0x01,0x00,
0x24,0x01,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x0e,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x27,0x06,0x00,0x01,0x00,0x26,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x12,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x27,0x04,0x00,0x00,
0x00,0x01,0x00,0x00,0x00,0x25,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x17,0x00,
0x00,0x00,0x00,0x00,0x00,0x26,0x04,0x00,
0x05,0x00,0x25,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x43,0x04,0x00,0x05,
0x00,0x25,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x43,0x32,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x44,0x33,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x44,0x10,
0x00,0x04,0x00,0x1c,0x00,0x00,0x00,0x33,
0x00,0x07,0x00,0x01,0x01,0x00,0x00,0x00,
0x00,0x01,0x10,0x00,0x04,0x00,0x05,0x00,
0x00,0x15,0x00,0x34,0x00,0x00,0x06,0x00,
0x01,0x00,0x00,0x00,0x00,0x00,0x02,0x10,
0x00,0x01,0x02,0x00,0x03,0x00,0x00,0x00,
0x00,0x4b,0x00,0x00,0x00,0x00,0x04,0x00,
0x01,0x00,0x00,0x00,0x00,0x00,0x02,0x10,
0x00,0x01,0x06,0x00,0x00,0x00,0x01,0x4b,
0x00,0x00,0x00,0x00,0x00,0x03,0x00,0x01,
0x00,0x00,0x00,0x00,0x03,0x10,0x00,0x01,
0x05,0x00,0x54,0x00,0x00,0x02,0x00,0x04,
0x00,0x00,0x02,0x10,0x00,0x01,0x05,0x00,
0x54,0x00,0x00,0x02,0x00,0x01,0x00,0x00,
0x00,0x00,0x00,0x02,0x10,0x00,0x01,0x06,
0x00,0x00,0x00,0x00,0x00,0x4b,0x01,0x00,
0x00,0x00,0x03,0x00,0x01,0x00,0x00,0x00,
0x00,0x00,0x02,0x10,0x00,0x01,0x07,0x00,
0x00,0x00,0x00,0x4b,0x00,0x00,0x00,0x00,
0x04,0x00,0x04,0x00,0x00,0x02,0x10,0x00,
0x01,0x09,0x00,0x00,0x4b,0x00,0x00,0x06,
0x00,0x03,0x00,0x00,0x03,0x1e,0x00,0x5f,
0x7f,0x7f,0x7f,};

unsigned missing_image_color_values[]={
0xffffff,0xffffff,0xffffff,0xffffff,0xffffff,0xa2a2a2,0x7f7f7f,0x9d9d9d,
0xe2e2e2,0xffffff,0xffffff,0x454545,0x000000,0x0c0c0c,0xdedede,0xffffff,
0xffffff,0x454545,0x000000,0x2b2b2b,0xa2a2a2,0x8d8d8d,0x040404,0x000000,
0x888888,0xffffff,0xffffff,0x454545,0x000000,0x3c3c3c,0xdddddd,0xd1d1d1,
0x111111,0x000000,0x9d9d9d,0xffffff,0xfdfefd,0xfefffe,0xffffff,0xffffff,
0x454545,0x000000,0x040404,0x737373,0xfbfbfb,0xffffff,0xfdfdfd,0xfefefe,
0xffffff,0xffffff,0x454545,0x000000,0x2b2b2b,0xa2a2a2,0x5d5d5d,0x000000,
0x333333,0xffffff,0xfefffe,0xfbfcfb,0xfdfdfd,0xffffff,0xffffff,0x454545,
0x000000,0x343434,0xc4c4c4,0x777777,0x000000,0xdddddd,0xffffff,0xfeffff,
0xf9fcfa,0xfcfcfc,0xfffeff,0xffffff,0xffffff,0x454545,0x000000,0x404040,
0xffffff,0xfefefe,0xfcfcfc,0xfefefe,0xffffff,0xfefffe,0xf8fbf8,0xf9fcfa,
0xfdfefd,0xffffff,0xfefffe,0xffffff,0xfefefe,0xfdfefd,0xffffff,0xffffff,
0xb7b7b7,0xa2a2a2,0xb3b3b3,0xfbfbfb,0xffffff,0xfafafa,0xeeede9,0xe1e0da,
0xd5d4c8,0xd8d7c8,0xe7e6dc,0xf6f5f2,0xffffff,0xf8fbf9,0xf5faf6,0xfdfdfd,
0xffffff,0xfefffe,0xffffff,0xfefefe,0xfbfcfc,0xfefffe,0xffffff,0xffffff,
0xfdfdfd,0xf5f5f3,0xe7e6e3,0xc6c4b8,0xa09e82,0x89875f,0x8e8b60,0xacaa89,
0xcbcab5,0xd9d8c9,0xe9e8df,0xfdfdfd,0xffffff,0xfafdfb,0xf6faf7,0xfbfefb,
0xffffff,0xfefefe,0xfefffe,0xfdfefd,0xfcfdfc,0xffffff,0xffffff,0xfbfafb,
0xe6e5e0,0xc5c3b5,0x93916f,0x716e38,0x747136,0x83824c,0x929161,0xabab84,
0xc2c2a5,0xd9d9c6,0xefefe6,0xfdfdfc,0xffffff,0xfcfdfc,0xf7faf8,0xfbfcfb,
0xfdfefd,0xfefefe,0xfcfdfc,0xfefeff,0xfdfefd,0xfcfdfd,0xffffff,0xffffff,
0xf6f5f5,0xd2d1c5,0x919067,0x767432,0x7b7a37,0x817f3e,0x858341,0x8d8b4d,
0x98975f,0xadad81,0xd8d7c2,0xf5f6f4,0xffffff,0xfcffff,0xfcfdff,0xfcfdfe,
0xfcfdff,0xfdfefe,0xfffffe,0xffffff,0xfdfefc,0xf7f9f6,0xfbfcfa,0xfdfffd,
0xfafdfb,0xf6f8f9,0xf8f7fb,0xfbfcfd,0xfdfefd,0xffffff,0xffffff,0x959595,
0x5d5d5d,0xffffff,0xf0efed,0xbdbba0,0x908c51,0x8c8842,0x8e8c49,0x8c8945,
0x898641,0x888540,0x918f50,0xa2a26e,0xbec1a3,0xd4dfda,0xddf2fe,0xdff4ff,
0xdef1ff,0xdff1fc,0xe2f1fc,0xe6f2fc,0xeaf4fd,0xeef7fd,0xf2f9fe,0xf7faff,
0xf9fcfd,0xf3f8f5,0xf8fbf8,0xfffffe,0xf8fbf7,0xeff3f4,0xedecf6,0xf4f4fa,
0xfbfdfc,0xfcfdfc,0xfefefe,0xffffff,0xffffff,0xdadada,0x000000,0xb5b5b5,
0xffffff,0xfefefd,0xfefffe,0xfeffff,0xf6fcfd,0xcad3cb,0x9e9d6b,0x9a9346,
0x9c9750,0x9a9752,0x989450,0x94904b,0x918e47,0x8f8b45,0x929254,0xa0a985,
0xaccbcf,0xb1defa,0xb1e1ff,0xb5e0fc,0xbae0f8,0xc0e1f8,0xc6e5f9,0xcde8fa,
0xd5eafb,0xddeffc,0xddeff5,0xe5f1f4,0xf3f9fb,0xf2f7f4,0xe9eced,0xe4e5f2,
0xececf6,0xfafafd,0xfdfefd,0xfefdfd,0xffffff,0xffffff,0xfafafa,0x383838,
0x000000,0x2a2a2a,0xfafafa,0xffffff,0xfafdff,0xeef7fb,0xe4f2fa,0xdceefc,
0xd3ecfd,0xc9ecff,0xabcedd,0x95a17a,0x9d984b,0xa29c50,0xa19c56,0xa09c55,
0x9f9b54,0x9c9951,0x98944c,0x948e43,0x939353,0x93a88c,0x8bbfd1,0x86cef8,
0x86d3ff,0x89cff7,0x8ed0f4,0x97d2f5,0xa0d5f6,0xaedbf7,0xb7dff3,0xc5e3f3,
0xd3ebf7,0xdaedf5,0xd5e3eb,0xd4dbed,0xdde2f1,0xf3f5f9,0xfffffd,0xfffefe,
0xffffff,0xffffff,0x868686,0x4a4a4a,0x000000,0x101010,0x292929,0xffffff,
0xfdfeff,0xf7fbff,0xebf5fe,0xddeffb,0xcde7f7,0xbce0f5,0xaad9f6,0x99d2f5,
0x8ccff4,0x84d1fe,0x86cbec,0xa3b997,0xb6b058,0xafaa55,0xa9a559,0xa8a358,
0xa6a358,0xa5a157,0xa29d56,0x9f994f,0x9b9244,0x939251,0x87a894,0x89c8de,
0x8fd6fd,0x91d8ff,0x93dafa,0x8ed6f8,0x84cff6,0x84ccef,0x93d0ed,0xa1d6f1,
0xb0dcf4,0xb6d6ec,0xb2cbe6,0xbdcfea,0xdae8f4,0xf5fafc,0xfefefd,0xffffff,
0xffffff,0xd1d1d1,0x131313,0x000000,0x060606,0x787878,0xffffff,0xfcfdff,
0xf3f9fe,0xe3f1fd,0xd3eafb,0xc4e4fa,0xafdbf6,0x93d1f2,0x81caf1,0x7fcbf2,
0x89d1f5,0xa2e0fa,0xb3e8fd,0xc8f2ff,0xcde8f8,0xb2baa4,0xb3b060,0xbbb85c,
0xb3af5e,0xb2ae5d,0xb0ac5c,0xada85a,0xaba75a,0xa8a45a,0xa59e51,0xa29848,
0xa9a361,0xbcc2a7,0xc0d0de,0xc0d6f6,0xc8e7fc,0xbdebfe,0xace5fa,0x9adaf0,
0x87cfee,0x7ec9ef,0x80c2e7,0x81b7e0,0x91bae1,0xb1d3ee,0xd4eaf8,0xe9f5fb,
0xf7fbfe,0xfefeff,0xffffff,0xffffff,0xfdfdfd,0x272727,0x3f3f3f,0x4e4e4e,
0x000000,0xffffff,0xfeffff,0xf5fafe,0xe5f2fc,0xd4ebfb,0xbfe1f9,0xa6d8f6,
0x88cdf4,0x78c8f0,0x7bcbf0,0xa1ddf8,0xc8f1ff,0xd8f5ff,0xd9ecf8,0xc0cce7,
0x9c9fcd,0x7e7dc0,0x5a58ac,0x71687b,0xc2bd61,0xcac962,0xbcb862,0xbbb861,
0xb9b560,0xb6b260,0xb3af5e,0xb0ad5d,0xaca85b,0xa69f4e,0xa39b4a,0x9a9567,
0x7c7898,0x7671ba,0x9ea0d4,0xc1d0e6,0xcde6ef,0xcbedf9,0xb7e8fc,0x8acceb,
0x61acdd,0x5ea6da,0x7cbee7,0xa9d7f4,0xc8e6f9,0xd9eefb,0xebf6fe,0xfafdfe,
0xffffff,0xffffff,0x000000,0xffffff,0xf8fbfe,0xe8f4fd,0xd5ecfb,0xc2e3f9,
0xa8d9f6,0x85ccf4,0x79c8f2,0x8cd1f3,0xafe1f6,0xd2f1fc,0xdbecf9,0xc0cbe4,
0x969ac9,0x6c68ad,0x4d3f94,0x38237f,0x311873,0x2d1872,0x2b1570,0x736466,
0xcfcc67,0xd1ce66,0xc6c265,0xc4c164,0xc2bf64,0xc0bc62,0xbdb962,0xbab761,
0xb7b360,0xb1ad5b,0xada750,0x979051,0x695b65,0x3a2780,0x463a95,0x7177b0,
0x9aa3cc,0xc1cde7,0xc4ddec,0x9ec7e3,0x7cb1dc,0x69b6e2,0x71c3ee,0x94d0f3,
0xb0dbf6,0xc6e5fa,0xdbeefc,0xf0f8fe,0xfdfdff,0xffffff,0xffffff,0x000000,
0xffffff,0xfdfeff,0xf2f8fe,0xddeffc,0xc7e5f9,0xaddbf7,0x8acef5,0x74c6f1,
0x8fd3f4,0xc1e8fa,0xe1f7fe,0xd9edf5,0xb2bedd,0x797bb6,0x4e4396,0x3b2583,
0x321876,0x2e126b,0x2d1465,0x2f1765,0x332267,0x2b1666,0x220a64,0x796b68,
0xdcd76a,0xdad56a,0xcfcb69,0xceca68,0xccc767,0xc9c567,0xc7c366,0xc3bf64,
0xbfbc63,0xbbb863,0xbbb85c,0xb3af55,0x7d6f5a,0x352269,0x2b247f,0x433f8f,
0x554d9c,0x767eb3,0x92a9cd,0x9eb6db,0xa6cae5,0x99d4ef,0x7dc9ef,0x77c6ef,
0x94d0f4,0xb4dcf7,0xcfe9fa,0xe6f3fd,0xf7fbff,0xffffff,0xffffff,0x000000,
0xffffff,0xf7fbfe,0xe6f3fd,0xd0e8fb,0xb6ddf9,0x93d1f5,0x79c9f3,0x8bd1f3,
0xb9e3f8,0xe4f7fe,0xddecf9,0xacbadb,0x7680b6,0x4f4897,0x3b2583,0x341875,
0x33196e,0x301968,0x301965,0x301c66,0x311d67,0x33216a,0x321e68,0x280f65,
0x270f65,0x84766a,0xe5e06e,0xe1dd6e,0xdad36d,0xd9d26b,0xd6d16b,0xd4ce6b,
0xd0cb6a,0xcdc968,0xc9c566,0xc4c065,0xc0bd61,0xbcba5b,0x99945a,0x57526a,
0x342e7e,0x32207c,0x3a2c83,0x45529a,0x556db0,0x849ecd,0xc0d9ee,0xd0eefa,
0xaadbf5,0x7bc6ee,0x77c3ee,0xa0d5f5,0xc1e3fa,0xd8ecfc,0xeef7fe,0xfbfdff,
0xffffff,0xffffff,0x000000,0xffffff,0xf1f8fd,0xdbeffc,0xc4e5f9,0xa7d7f7,
0x80c9f3,0x78c9f2,0xaedef7,0xe2f3fc,0xeaf8ff,0xbdcee8,0x7e89c0,0x54509b,
0x423288,0x3c2581,0x392079,0x321b6d,0x301b66,0x301a65,0x321d68,0x311c68,
0x331f69,0x311c67,0x311965,0x291066,0x281166,0x90846d,0xece872,0xe6e071,
0xe1dc70,0xe0db6f,0xded96f,0xdcd76d,0xdbd56d,0xd8d16b,0xd3cd69,0xccc869,
0xc6c165,0xc5c15e,0xaeaa5b,0x6e6d6b,0x302774,0x2e1c75,0x353586,0x2e4092,
0x3e56a0,0x7087c0,0xbdd0e9,0xe5f4fe,0xcae3f7,0x8ec9ed,0x72c2ef,0x8ecef4,
0xb1dcf7,0xcce8fa,0xe4f3fd,0xf9fcfe,0xffffff,0xffffff,0x000000,0xffffff,
0xfeffff,0xf1f8fd,0xd7ecfb,0xbce1f9,0x99d3f6,0x76c7f1,0x8cd1f3,0xd0ebfa,
0xeef9ff,0xe0f2fc,0xabbfe2,0x6a70b0,0x4d4193,0x422f87,0x412c86,0x3e2681,
0x372074,0x311a6b,0x311c67,0x301a65,0x301965,0x321e67,0x311d67,0x321d68,
0x311b66,0x301864,0x301965,0x280f64,0x2f1866,0xb0a56f,0xf8f474,0xe8e372,
0xe7e172,0xe5e071,0xe3df70,0xe2dd70,0xe0db70,0xded86e,0xd9d46c,0xd2cc6a,
0xcac566,0xc5c160,0xb9b75d,0x6f6864,0x2c1f71,0x2e2d83,0x314090,0x314797,
0x3e4b99,0x676faf,0xa5b8dd,0xd0e3f5,0xd7e4f5,0xabd6f2,0x77c8f1,0x7fcaf3,
0xa7d9f7,0xc6e5fa,0xe1f1fc,0xf6fbfe,0xffffff,0xffffff,0x000000,0xffffff,
0xfdfeff,0xecf6fe,0xd4ebfb,0xb4def8,0x8acdf3,0x73c7f1,0xa4dbf5,0xe3f2fb,
0xeff9ff,0xd6ecfb,0x9eb2da,0x6c6fb0,0x4f4395,0x453089,0x432f86,0x432e87,
0x3f2783,0x3b217b,0x381d74,0x372072,0x341d6d,0x321a6a,0x33206b,0x321e69,
0x331e6a,0x311a68,0x311866,0x311966,0x311a66,0x250d65,0x453169,0xd5ce74,
0xfbf775,0xece575,0xece674,0xebe674,0xeae573,0xe8e372,0xe6e071,0xe3de71,
0xdfd96f,0xd7d26b,0xcdc868,0xc6c363,0xbbb75c,0x767067,0x303283,0x293891,
0x324d99,0x364494,0x464292,0x6263a9,0x88a0cf,0xbfd4ee,0xdfeaf9,0xcee8f9,
0x8bd1f3,0x76c7f2,0x9bd4f5,0xc2e3f9,0xdceffc,0xf4faff,0xffffff,0xffffff,
0x000000,0xffffff,0xfdfeff,0xebf6fe,0xd1eafb,0xb3ddf7,0x85cdf3,0x7ccaf2,
0xb5e1f7,0xe9f4fb,0xe9f6fe,0xd5ecfb,0xa3b8de,0x7279b7,0x5c56a0,0x4e3f92,
0x46338a,0x46378a,0x433287,0x402983,0x3f2382,0x3e2280,0x3e2781,0x3c257f,
0x3b227d,0x3d297e,0x3c277c,0x3d277c,0x3a2279,0x3a2079,0x392078,0x382077,
0x2a1077,0x6f5d79,0xf0ea78,0xf9f575,0xf0ea77,0xf0eb76,0xf0ea76,0xefe875,
0xede774,0xebe574,0xe7e373,0xe3dd70,0xdbd66d,0xd0cb6a,0xc8c363,0xbcba5e,
0x767672,0x2f3c93,0x29489d,0x354996,0x414191,0x504998,0x5e67ab,0x8399cc,
0xbfd6f0,0xe8f3fe,0xddf0fc,0x9bd7f4,0x79c9f1,0x96d3f6,0xbfe2f9,0xdbeefb,
0xf3f9fe,0xffffff,0xffffff,0x000000,0xffffff,0xfdfeff,0xe9f6fc,0xcee8fa,
0xabdaf7,0x7bc9f2,0x88d0f3,0xd0ebf9,0xebf4fc,0xe6f5fc,0xd6eefb,0xa9c3e5,
0x818dc3,0x6a6cae,0x5f5ba3,0x564c99,0x514696,0x534e97,0x4f4894,0x4b3d8f,
0x48368d,0x46338b,0x48388e,0x46348c,0x45318a,0x46358c,0x443089,0x443088,
0x412a86,0x402786,0x412786,0x402686,0x412686,0x3d2386,0x331885,0x8d7e81,
0xfefc79,0xf9f478,0xf4ed78,0xf4ee78,0xf3ed77,0xf2ec77,0xefea76,0xebe574,
0xe6e171,0xdfda6f,0xd5cf6b,0xcdc965,0xc3bd5f,0x838673,0x2f4d9c,0x304b9f,
0x414794,0x4e4796,0x5557a1,0x5f6caf,0x8ca4d2,0xcbe3f7,0xe9f6ff,0xe9f4fc,
0xb4e1f6,0x75c7f2,0x8dcff4,0xbadff8,0xd8ecfc,0xf4f9fe,0xffffff,0xffffff,
0x000000,0xffffff,0xfeffff,0xeaf5fd,0xcde8fa,0xacdaf7,0x7dcaf2,0x85cef3,
0xd5edfa,0xecf5fc,0xe3f1fb,0xe1f4fe,0xc4e1f6,0x9bb2db,0x8792c8,0x7a81bc,
0x7175b4,0x6969ac,0x6463a8,0x6267a6,0x5d5fa2,0x56509b,0x504495,0x4b3e91,
0x4a3d8f,0x46358b,0x433088,0x443289,0x432b87,0x422c87,0x402684,0x3f2584,
0x402484,0x3f2483,0x402683,0x351b84,0x3b2284,0xb7ac7f,0xffff7a,0xfaf27a,
0xf6ef79,0xf7f079,0xf5f079,0xf5ef79,0xf0eb76,0xece775,0xe8e373,0xe2dc71,
0xd8d26c,0xd0cb66,0xcac35d,0x7a887f,0x3655a8,0x4852a0,0x58579f,0x595fa6,
0x5d68ac,0x7486c0,0xa8c3e6,0xd7eefd,0xe8f5fd,0xebf5fb,0xb6e1f7,0x77c9f2,
0x8bcdf4,0xbadff8,0xdaedfc,0xf6fafe,0xffffff,0xffffff,0x000000,0xffffff,
0xfffeff,0xedf7fe,0xcee9fa,0xacdaf6,0x78c8f2,0x8ad1f3,0xd7eefa,0xebf5fc,
0xe1f1fb,0xe2f2fc,0xd4eefd,0xb9d7f1,0xa0b6dc,0x90a0cf,0x828ec4,0x767fb9,
0x6b6cae,0x6462a7,0x5e5fa2,0x55519b,0x4d4091,0x45348b,0x432b89,0x412c87,
0x3f2883,0x3e2683,0x402b84,0x402682,0x402981,0x3c237e,0x3a227a,0x392277,
0x392076,0x382177,0x271177,0x4a357a,0xd9d37a,0xffff7a,0xf9f279,0xf9f27a,
0xf8f27a,0xf8f179,0xfbf57b,0xfbf47c,0xf3ec77,0xe9e473,0xe2de70,0xd9d56d,
0xd5cf66,0xc8c563,0x77868e,0x5062b2,0x6c76b3,0x6d7bb6,0x6979b8,0x788bc3,
0x98b1da,0xc3e0f6,0xe1f3fd,0xe5f2fc,0xeef6fc,0xbce5f8,0x74c8f1,0x8acdf3,
0xbce0f8,0xdceffb,0xf6faff,0xffffff,0xffffff,0x000000,0xffffff,0xfeffff,
0xf3f9fe,0xd5ebfc,0xb1ddf7,0x80caf2,0x81cdf2,0xd3ecfa,0xedf6fc,0xe2f1fb,
0xe2f2fc,0xd5edfc,0xbcdbf4,0xa5bde1,0x90a1cf,0x818bc2,0x7176b4,0x6565a9,
0x5a529e,0x524897,0x504894,0x4b3c8e,0x422985,0x412d87,0x423287,0x3f2b87,
0x3f2888,0x3c2785,0x3c277f,0x382075,0x382373,0x341e6f,0x321c6b,0x321d6b,
0x311e6a,0x311e69,0x321f69,0x2d1a68,0x1e086a,0x7b6f73,0xf9f879,0xfefb79,
0xfaf278,0xfaf379,0xfcf67a,0xf9f37a,0xf5f079,0xfff97e,0xfffc7e,0xfcf57a,
0xf0e975,0xded86f,0xd9d365,0xc6c369,0x78829b,0x6571bf,0x7285c6,0x7087c9,
0x8199ce,0x9db8de,0xbcdcf5,0xddf1fc,0xe2f2fb,0xe4f2fb,0xecf5fc,0xb7e2f6,
0x75c8f1,0x93d1f4,0xc2e3f9,0xe0f0fd,0xf9fcfe,0xffffff,0xffffff,0x000000,
0xffffff,0xf6fbfe,0xdceefc,0xbadff9,0x83ccf3,0x81cef2,0xd1ecf9,0xecf5fc,
0xe2f1fb,0xe1f1fb,0xd4edfc,0xbad8f1,0x9fb6dd,0x8a98ca,0x797fbb,0x6a6bae,
0x5d58a1,0x544998,0x4b3b90,0x46398c,0x48428f,0x483e8d,0x43368f,0x3a3a77,
0x2a3c41,0x27363e,0x2c3455,0x303466,0x333d70,0x36427a,0x37467b,0x344578,
0x344377,0x343e76,0x333b76,0x333974,0x333572,0x322f70,0x332a6c,0x271969,
0x2c216e,0xb6b378,0xffff79,0xfcf478,0xfaf378,0xfefa7b,0xf0f077,0x707841,
0x414c2e,0x666d3e,0x959852,0xbbbb62,0xd1cd68,0xeae172,0xe2dc70,0xc7c260,
0x9d9b68,0x64688c,0x677296,0x717e9d,0x7c91be,0x91acdb,0xafcdec,0xcce9fb,
0xdef2fc,0xe1f1fb,0xe4f2fb,0xeef6fc,0xb2e0f6,0x76c8f2,0x9cd5f6,0xc6e5fa,
0xe5f2fc,0xfdfdff,0xffffff,0xffffff,0x000000,0xffffff,0xfbfdff,0xe1f1fc,
0xc2e3f9,0x90d0f4,0x79caf3,0xc5e8f9,0xeef5fc,0xe4f2fb,0xe1f2fc,0xd3ecfb,
0xb7d5ef,0x9db3da,0x8693c6,0x747ab7,0x6463a8,0x59559e,0x4f4e98,0x484c94,
0x424c90,0x425591,0x486796,0x4a6d9b,0x437098,0x2b5255,0x1b3b1a,0x1c3c1d,
0x1d3a22,0x1d3a25,0x147745,0x308080,0x41688d,0x396685,0x3a5b83,0x37527f,
0x35487c,0x353f77,0x343874,0x323173,0x322b6f,0x32266c,0x1b156f,0x565576,
0xe5e47a,0xfffe79,0xfdf57a,0xffff7f,0xb0b65e,0x10221b,0x041515,0x0a1b17,
0x11211a,0x1b261e,0x286c38,0xbbdf6c,0xf4e976,0x989659,0x868750,0x838452,
0xa5a35c,0xbab55c,0x8f9796,0x748ccb,0x92abd8,0xb1ceed,0xcce8fb,0xe0f1fd,
0xe1f1fb,0xe6f2fc,0xe8f4fc,0xa1daf5,0x78c8f2,0xa4d7f6,0xcce7fa,0xedf5fd,
0xfefeff,0xffffff,0xffffff,0x000000,0xffffff,0xfdfeff,0xeaf4fc,0xcae6fa,
0x9dd4f5,0x7acaf2,0xbbe3f8,0xeef6fc,0xe3f2fb,0xd6eefc,0xb6d3f0,0x99aed8,
0x838fc4,0x7076b4,0x6160a5,0x55519b,0x4c5096,0x475495,0x435b95,0x426496,
0x426d97,0x497797,0x4d7d9e,0x3a6f7e,0x1f472f,0x1b3b1d,0x1e3f25,0x203824,
0x184d27,0x00a942,0x1d9869,0x43708d,0x3a6b87,0x3a6284,0x385981,0x35507e,
0x354679,0x343e76,0x333775,0x333071,0x312a6e,0x2b2670,0x1d2176,0x999879,
0xffff79,0xfffe7b,0xf7f77b,0x586539,0x071a18,0x192b1f,0x15281e,0x15211b,
0x112a1e,0x00873e,0x54ca5d,0xebe474,0x9d995a,0xa7a55c,0xb3af5c,0xafac61,
0xaaa660,0x8c8f8f,0x6273ba,0x7a8dc6,0x96abd6,0xb1ceeb,0xcfeafc,0xe1f1fc,
0xe2f1fb,0xe7f3fb,0xe5f2fc,0x90d2f4,0x7ac8f2,0xafdbf7,0xd5ebfa,0xf3f9fe,
0xffffff,0xffffff,0x000000,0xffffff,0xf3f9fe,0xd5ebfb,0xaddaf7,0x7bc9f2,
0x9cd8f5,0xe8f4fc,0xe6f3fc,0xe2f1fb,0xd9effb,0xbbd9f2,0x99b0d9,0x8190c3,
0x6e77b3,0x6062a6,0x54539b,0x4a4f95,0x455493,0x425b94,0x416395,0x3f6d93,
0x407691,0x488393,0x508c98,0x306358,0x193c1c,0x1d4124,0x1d4026,0x203723,
0x107337,0x00b848,0x0ca954,0x398187,0x3e728e,0x396c88,0x386286,0x365883,
0x37507f,0x36477c,0x343c78,0x333373,0x312a6e,0x302d71,0x212677,0x403f75,
0xd5d178,0xffff7e,0xced06a,0x203323,0x12261d,0x192d1f,0x192c1f,0x1d221c,
0x125c32,0x00ad4f,0x0fb250,0xc6df6e,0xf2e676,0xe4df72,0xe5de6c,0xc1bc6c,
0x696988,0x4c509e,0x5862aa,0x6b77b6,0x828fc5,0x99aed7,0xb6d5f0,0xd4ecfc,
0xe3f2fc,0xe1f1fb,0xe9f4fc,0xd1ecfa,0x7ecbf3,0x8bcef4,0xbee1f9,0xe1f1fd,
0xfbfdff,0xffffff,0xffffff,0x000000,0xffffff,0xfcfdff,0xe1f1fd,0xbee1f9,
0x88ccf4,0x7fccf2,0xd4ecf9,0xe5f2fa,0xddedf8,0xd7ecf8,0xbad9ef,0x97b4d7,
0x7992bf,0x657db0,0x566da5,0x4a639c,0x436397,0x416b96,0x427296,0x427897,
0x407b94,0x3d7d8d,0x40868b,0x4d9397,0x4f8b86,0x245234,0x193e1f,0x1e4528,
0x1f3d25,0x1b4a29,0x059e45,0x00b44a,0x02b048,0x248871,0x41618e,0x395e84,
0x375481,0x384c7e,0x36437a,0x343976,0x333071,0x312a6f,0x31276d,0x2e2f74,
0x2e377a,0x1a1e76,0x827d78,0xfaf77e,0x748145,0x0b221b,0x1a2f21,0x192e21,
0x1c261e,0x1a3924,0x079145,0x00b552,0x00a74b,0x72c95f,0xfff37a,0xeee775,
0xe3de70,0xe4dd67,0xa19e75,0x424296,0x454da2,0x5d64a9,0x6f77b4,0x8390c4,
0x9eb3db,0xbcdaf3,0xdaf0fe,0xe4f2fc,0xe0f1fb,0xeaf4fc,0xc0e5f9,0x76c9f2,
0x9dd5f5,0xcbe6fb,0xeef7fd,0xffffff,0xffffff,0x000000,0xffffff,0xedf7fd,
0xcbe7fa,0x9dd5f6,0x78c9f2,0xb6def4,0xdbeaf2,0xcce2ec,0xc2dde9,0xa8d0e2,
0x86b1ce,0x6894b8,0x5181a9,0x4374a1,0x366c9a,0x2d6995,0x2c6b93,0x2e6e92,
0x316e92,0x336b8e,0x356788,0x346886,0x39718a,0x467792,0x39635e,0x19451f,
0x1d4625,0x1d4626,0x1f3c24,0x0f7736,0x00b44c,0x00ae4a,0x00b246,0x0f9a59,
0x365f82,0x3a4f85,0x364d7e,0x35467b,0x344079,0x323976,0x333474,0x313474,
0x303474,0x2d3b7b,0x2a4380,0x23357c,0x424880,0x949c69,0x2a3e23,0x162c20,
0x1c3122,0x1b2f21,0x1e271e,0x126333,0x01b050,0x00b050,0x00a74c,0x26b353,
0xdfe874,0xfaee78,0xe5e273,0xe4df6d,0xd7d367,0x6f6b81,0x353e9d,0x4d54a0,
0x6160a6,0x7076b4,0x8794c7,0xa3badf,0xc2e1f7,0xddf1fd,0xe3f1fc,0xe4f1fb,
0xe8f4fc,0x9dd8f5,0x7ac9f2,0xb1dcf8,0xdaeefb,0xf7fbff,0xffffff,0xffffff,
0x000000,0xffffff,0xf9fbff,0xdaedfc,0xb2ddf7,0x7bc9f1,0x93d1ed,0xccdfe7,
0xb6d2dc,0xaaccd8,0x90c1d0,0x6ea8bf,0x5291af,0x4082a5,0x31759d,0x216c97,
0x156692,0x0f628f,0x0d628e,0x0b638a,0x0c6485,0x166482,0x206980,0x277081,
0x367c8b,0x3a7179,0x224d2f,0x1c4a21,0x1e4c28,0x1f4525,0x1a5129,0x05a144,
0x00b34c,0x00ad4a,0x00b048,0x01af48,0x1d886e,0x316989,0x286883,0x286181,
0x2a5a80,0x2b537e,0x2d4c7e,0x2f4a7e,0x32487f,0x304c83,0x2e4f88,0x304885,
0x2f3f82,0x233b42,0x1a331b,0x1d3421,0x1d3322,0x1e2d20,0x1c3924,0x088f43,
0x00b552,0x00ab4e,0x00aa4e,0x03a84e,0x95d468,0xfff47b,0xede875,0xe3df72,
0xe7e269,0xb3ac6e,0x3d4995,0x38439d,0x534d9a,0x615fa5,0x7379b7,0x8b99ca,
0xa8c3e4,0xcae6f9,0xe0f2fd,0xe3f1fb,0xe6f3fb,0xd8eefa,0x7ecbf1,0x8ecff4,
0xc6e5f9,0xe6f4fd,0xfdffff,0xffffff,0xffffff,0x000000,0xffffff,0xfeffff,
0xeaf5fd,0xc6e5fa,0x94d1f4,0x75c6ed,0xbadbe8,0xbbd2dd,0xa6c8d5,0x92c1d2,
0x71aec3,0x5395b0,0x3e83a4,0x32779d,0x256c96,0x156891,0x106b8e,0x0e718d,
0x0e768a,0x0f7886,0x167c84,0x288184,0x398985,0x469188,0x529b8d,0x367a57,
0x194b1e,0x1e5028,0x1e4f28,0x204424,0x0f7a36,0x00b34c,0x00af4a,0x00ad49,
0x00b145,0x10a85c,0x32888b,0x337e91,0x347f8e,0x387b8f,0x3c7891,0x3d7592,
0x417294,0x487296,0x4b759a,0x49759e,0x4a70a4,0x3c5779,0x193326,0x1c3520,
0x1d3624,0x1d3422,0x202e20,0x136434,0x01b050,0x00b050,0x00aa4e,0x00ab4f,
0x00a74d,0x39b857,0xeded77,0xf9ee78,0xe8e473,0xe4df6f,0xdfd767,0x707a85,
0x2e3d9a,0x464094,0x554e9b,0x6564a9,0x7881bb,0x92a3d0,0xb1ceec,0xd3edfc,
0xe2f2fb,0xe2f1fb,0xeaf5fb,0xb6e1f7,0x7bc9f3,0xabdaf7,0xd3ecfa,0xf5fbfe,
0xffffff,0xffffff,0x000000,0xffffff,0xf9fcff,0xdaeefd,0xb0dcf9,0x78c8f2,
0xa1d7f2,0xd7e7ef,0xc6dde7,0xb6d7e3,0x99c7d6,0x7bacc0,0x6191ae,0x527da5,
0x4c71a1,0x4a699d,0x476999,0x446898,0x3d6692,0x346089,0x2e5e82,0x2f6381,
0x336b83,0x377383,0x3c7d88,0x3a7e7d,0x205c34,0x185022,0x1c5428,0x1f4c26,
0x1a582b,0x059f44,0x00b34c,0x00ad4a,0x00af47,0x09ab4f,0x44957a,0x5d8c8a,
0x588d87,0x5a8a87,0x5c8989,0x5c8a8b,0x618890,0x648793,0x417494,0x266897,
0x2b679a,0x294d4f,0x203b1f,0x1d3924,0x1d3925,0x1e3122,0x1c4127,0x089346,
0x00b553,0x00aa4e,0x00ab4e,0x00a94f,0x09a94f,0xa6d969,0xfff57d,0xede875,
0xe5e073,0xe3dc6a,0xb7b770,0x424f92,0x383593,0x4a4395,0x59539e,0x6b6cae,
0x808ac1,0x9bb2da,0xbfdef5,0xdaf0fd,0xe3f1fb,0xe6f4fb,0xdff0fb,0x8ad0f3,
0x8ccff4,0xc2e3f9,0xe7f3fd,0xfefffe,0xffffff,0xffffff,0x000000,0xffffff,
0xfeffff,0xecf6fd,0xcae5f8,0x99d2f2,0x85cdef,0xd4eaf2,0xe0eef3,0xd8ebf4,
0xc2dfed,0xa6c4da,0x8aa1c2,0x7183b0,0x6171a3,0x586b9b,0x4c6795,0x416691,
0x36658f,0x2c6786,0x2a6e7e,0x327979,0x318278,0x2e8a78,0x269078,0x23947d,
0x1d7553,0x145220,0x165626,0x185526,0x1c4d24,0x0e7f38,0x00b24c,0x00b04b,
0x00ad4a,0x02b247,0x189765,0x257485,0x276e84,0x2c6b82,0x306581,0x326483,
0x336587,0x346286,0x205888,0x105593,0x154e70,0x1b3d26,0x1f3d21,0x1f3e26,
0x1f3a25,0x203322,0x136635,0x00af51,0x00ae50,0x00ab4e,0x00aa50,0x00a54c,
0x52c05b,0xfdf27b,0xf7ed78,0xe8e373,0xe0da6f,0xdbd766,0x777b81,0x313091,
0x413b8f,0x4f4595,0x5e59a2,0x7075b3,0x8a98ca,0xabc6e7,0xcae7fa,0xddf1fd,
0xe4f2fb,0xe9f4fc,0xb1e0f7,0x78c8f2,0xabd9f7,0xd6ebfb,0xf6fbfe,0xffffff,
0xffffff,0x000000,0xffffff,0xfefefd,0xfdfdf8,0xf8f9f7,0xddeef8,0xb9dff3,
0x88cdf0,0xb1ddf2,0xe5efef,0xd7e8e7,0xc1dddf,0x9bc3cf,0x76a5ba,0x558fa7,
0x3a7f9a,0x2a7891,0x24798c,0x1a7a88,0x137a88,0x117c84,0x0e7e80,0x0f827e,
0x13867b,0x168a78,0x198e77,0x1b8f7c,0x14856c,0x0e622e,0x135923,0x145b28,
0x185323,0x14612a,0x04a245,0x00b44c,0x00ad4a,0x00ac4c,0x00b247,0x00a253,
0x04727d,0x076689,0x076986,0x096685,0x0b6689,0x0b678c,0x0d6488,0x0e6891,
0x0f6991,0x194c45,0x1f3c1c,0x1e3f25,0x1f3f26,0x203923,0x1c4528,0x079547,
0x00b453,0x00aa4e,0x00ab4e,0x00ab4f,0x00aa50,0x00a64e,0x13ac4e,0xb5dd6d,
0xfff47c,0xeee775,0xe3dd71,0xe0da69,0xb2af6f,0x3f418f,0x373590,0x463a8f,
0x534998,0x6361a7,0x7980bb,0x97abd6,0xb6d4ef,0xceebfc,0xdff0fc,0xe8f3fc,
0xd8eefb,0x85cef3,0x8acef4,0xc5e4f9,0xe9f4fc,0xfdfeff,0xffffff,0xffffff,
0x000000,0xffffff,0xfffffe,0xfefefc,0xfbfcf8,0xfbfbf6,0xfcfcf6,0xf0f8f9,
0xcce7f2,0x9ad2e7,0x73c4e0,0xadd7da,0xb7d6cc,0x98c7be,0x6cb4b1,0x3f9fa0,
0x1e9391,0x128f87,0x128f81,0x128e7f,0x0f8c7e,0x0b8c7d,0x0d8b7c,0x0d897b,
0x0f867b,0x15847c,0x19847c,0x23847c,0x31857f,0x368285,0x196c4d,0x085c1e,
0x0e5e26,0x115c27,0x135524,0x098237,0x00b24b,0x00af4a,0x00ad4a,0x00ad4b,
0x00ad4a,0x00b145,0x0c9a66,0x1d8785,0x1b8883,0x1a8883,0x1a8a8a,0x0f888f,
0x007f8c,0x008599,0x087075,0x1b4528,0x1f3f21,0x1e4226,0x1e4026,0x203b23,
0x136834,0x00b152,0x00ad50,0x00ab4e,0x00ab4f,0x00ab50,0x00a94f,0x00a34a,
0x5bc15c,0xfdf27a,0xf8ed78,0xe6e272,0xded96e,0xd4ce67,0x626a84,0x2c3495,
0x3f338a,0x4a3b8f,0x59529d,0x6c6eaf,0x8794c7,0xa4bde0,0xbdddf5,0xd7edfc,
0xe5f2fb,0xeaf4fb,0xa8dcf6,0x75c7f1,0xb1dcf7,0xdaedfc,0xfafdff,0xffffff,
0xffffff,0x000000,0xffffff,0xfdfdf9,0xfbfbf6,0xfbfcf8,0xfefefa,0xf7f9f7,
0xd3e7ed,0x9fcedd,0x5fb7cc,0x5fb4bf,0x7db9b3,0x6ab2a6,0x43a99a,0x1b9e8f,
0x0a9584,0x0c947c,0x149578,0x169478,0x139279,0x0c9077,0x098e77,0x0a8d76,
0x0c8b77,0x158d77,0x1a8e77,0x1e9075,0x229374,0x279a79,0x23936f,0x066a2e,
0x016222,0x066324,0x0a5a22,0x0b6528,0x02a144,0x00b44d,0x00ad49,0x00ad4a,
0x00ad4c,0x00ac4c,0x00af47,0x08a952,0x1c8f76,0x1e8a7c,0x1b8c7b,0x198f82,
0x119087,0x088d88,0x078888,0x125a48,0x1e401e,0x1e4326,0x1d4327,0x1f3e24,
0x1a4d29,0x059a48,0x00b253,0x00aa4f,0x00ab4e,0x00ab4f,0x00aa51,0x00aa50,
0x00a44b,0x1fad50,0xc5e16f,0xfff37b,0xe9e574,0xe1db70,0xded768,0x949875,
0x303e93,0x392f8a,0x443188,0x504495,0x615ca5,0x7980bb,0x94a6d2,0xacc9e9,
0xc9e5fa,0xdff1fc,0xebf4fc,0xd2ecfa,0x84cef3,0x94d1f4,0xcbe7fa,0xf1f8fe,
0xffffff,0xffffff,0x000000,0xffffff,0xfefefd,0xfefefb,0xffffff,0xe8eff4,
0xaecfe0,0x6db4cd,0x36a5bd,0x59aaaf,0x64a7a6,0x44a0a0,0x1b9999,0x0a9491,
0x0f9288,0x149282,0x16917f,0x128e7f,0x108d80,0x168f7e,0x1e907a,0x219277,
0x219275,0x209277,0x1b9276,0x199072,0x199172,0x189878,0x0d8456,0x006623,
0x016625,0x006525,0x015d21,0x028133,0x00b24c,0x00af4c,0x00ad49,0x00ad4a,
0x00ad4c,0x00ac4c,0x00ae4a,0x00ae48,0x0f966a,0x0f857e,0x0b877c,0x088c82,
0x098f87,0x0b938b,0x137b6d,0x1e4929,0x1d4220,0x1d4627,0x1d4327,0x1f3e24,
0x117136,0x00b152,0x00ac50,0x00ac4e,0x00ab4e,0x00ab4f,0x00aa51,0x00aa4f,
0x00a84d,0x00a44a,0x7dcb62,0xfef47c,0xf1e976,0xe3df71,0xded86c,0xc4c268,
0x4c5a88,0x322f8e,0x412a85,0x49398e,0x58509d,0x6d70b0,0x8693c7,0x9eb3db,
0xb9d7f2,0xd5edfd,0xe6f3fc,0xe6f3fb,0x9dd8f5,0x81cbf3,0xbbe1f9,0xe7f3fd,
0xffffff,0xffffff,0x000000,0xffffff,0xfffffe,0xfdfdfc,0xd6e5e9,0xa1c8d3,
0x6bb4c7,0x56adc0,0x83b7bb,0x87b7b5,0x73b1ad,0x6aaaa2,0x5ca19a,0x479694,
0x328891,0x217f8e,0x197b8a,0x197d87,0x1f7f82,0x23817d,0x22817c,0x1f807c,
0x187d7d,0x0f7a7e,0x0c797f,0x0c7a84,0x107e7e,0x09723d,0x006a24,0x016928,
0x016324,0x016926,0x009e41,0x00b44c,0x00ad4a,0x00ad49,0x00ad4a,0x00ad4b,
0x00b047,0x01a15a,0x0f8c80,0x108c80,0x0c8e87,0x0b9193,0x0d8c8e,0x15624b,
0x1f451e,0x1d4826,0x1d4726,0x1d4325,0x1c4826,0x079746,0x00b254,0x00ab50,
0x00ac4e,0x00ab4e,0x00ab50,0x00ab51,0x00a94f,0x00a94e,0x00a34a,0x31b454,
0xe1ea75,0xfff17a,0xe6e372,0xe0db6e,0xd6d266,0x707b7b,0x28288a,0x391d7f,
0x3a2383,0x493d91,0x605fa7,0x7982bd,0x91a0cf,0xaac5e5,0xc8e6fa,0xdff1fd,
0xedf5fc,0xbde4f8,0x7dcaf3,0xa7d9f7,0xdbeffc,0xf9fcff,0xffffff,0xffffff,
0x000000,0xffffff,0xfffffe,0xfefefd,0xfdfefc,0xfffefe,0xfafcfa,0xd9e7e4,
0xb5d4d4,0x84c4d6,0x9acedf,0xcde0e6,0xc0dae1,0xa7c8d0,0x90abba,0x7691ae,
0x5a7aa2,0x3f6899,0x275f92,0x1e5f8e,0x1b648e,0x136689,0x0e6985,0x0d6c86,
0x096d86,0x077085,0x087785,0x077b84,0x0b7f8b,0x047561,0x006d2b,0x006f28,
0x006c2b,0x006527,0x018333,0x00b04a,0x00af4b,0x00ad49,0x00ad4a,0x00ad49,
0x00ad4a,0x00ad4c,0x00ad4b,0x00ad49,0x03ae4b,0x3ca477,0x579c90,0x489b96,
0x47a0a7,0x418a83,0x22542c,0x1b4a21,0x1e4e28,0x1e4b27,0x204324,0x156832,
0x01ac50,0x00ac51,0x00ab50,0x00ab4e,0x00aa50,0x00aa51,0x00aa4f,0x00a94e,
0x00a64d,0x07a54c,0x8fd164,0xfdf27a,0xf0e876,0xe3dd70,0xded868,0x9da272,
0x38438d,0x503c8c,0x57428d,0x554992,0x5f5ba0,0x7074b2,0x838ec4,0x9bb1d9,
0xb8d7f1,0xd5edfc,0xecf5fc,0xd7eefb,0x85cef3,0x92d1f4,0xcee8fb,0xf1f8fe,
0xffffff,0xffffff,0x000000,0xffffff,0xfffffe,0xfefffe,0xffffff,0xfafdfe,
0xdbecf5,0xacd8ee,0x81cbed,0xbfe3f1,0xe9f3f1,0xc9e5e9,0xa6c5db,0x88a2c6,
0x738db6,0x6782aa,0x607ea1,0x557b98,0x517e96,0x4f8391,0x46838b,0x3c8686,
0x3c8c82,0x39907e,0x36957b,0x3d9b77,0x44a27b,0x389d78,0x0c7b39,0x007026,
0x00722d,0x006d2c,0x00732f,0x00a244,0x00b24a,0x00ad49,0x00ad4a,0x00ad49,
0x00ad4a,0x00ad4b,0x00ac4a,0x00ad4a,0x00ac4c,0x00ae47,0x21a75f,0x62988a,
0x639c97,0x57999d,0x346c55,0x1b4c1e,0x1d5028,0x1e4c26,0x1b5127,0x069744,
0x00b153,0x00ab50,0x00ac4e,0x00ab4e,0x00ab50,0x00aa51,0x00aa4e,0x00a84e,
0x00a149,0x38b454,0xe2e974,0xfdee79,0xe5e172,0xe0d96a,0xbdc170,0x6f829d,
0x7c73a6,0x8675a7,0x867aac,0x6f67a4,0x6463a7,0x757db9,0x8e9dce,0xaac4e5,
0xc9e6f9,0xe4f3fd,0xe8f4fc,0x9bd7f5,0x82cbf3,0xc0e2f9,0xe8f4fe,0xffffff,
0xffffff,0x000000,0xffffff,0xfffeff,0xfefffd,0xfffffd,0xfdfdf8,0xf2f7f7,
0xd3eafa,0xa3d6f3,0x9bd2e8,0xd9eae7,0xe3efe7,0xbedbdb,0xa4bec7,0x90a9b3,
0x83a0a7,0x789a9f,0x6a9298,0x5a8d91,0x508e8d,0x509085,0x50937f,0x53977a,
0x559a75,0x589d73,0x539e71,0x519f6e,0x59a377,0x31905c,0x02752f,0x00732d,
0x00732e,0x006e2b,0x008a39,0x00b34a,0x00ae48,0x00ad4a,0x00ad49,0x00ad4a,
0x00ac4c,0x00ac4b,0x00ad4a,0x00ac4c,0x00ac4a,0x04ad4b,0x389c71,0x459692,
0x348377,0x155527,0x185123,0x1d5229,0x1d5228,0x204a24,0x107234,0x00b04f,
0x00ad51,0x00ab51,0x00ab50,0x00ab4e,0x00ab4f,0x00aa51,0x00aa4e,0x00a94e,
0x00a84e,0x00a54d,0x0aa74c,0xafda6a,0xfff179,0xebe573,0xe4dd6e,0xcccb6d,
0x727987,0x493b88,0x4b318b,0x574194,0x584a97,0x5c58a1,0x6d70b0,0x848fc4,
0x9fb4db,0xbedcf3,0xdcf1fd,0xeff6fd,0xaddef7,0x79c8f2,0xb2ddf7,0xe0f0fc,
0xfcfdff,0xffffff,0xffffff,0x000000,0xffffff,0xfefefe,0xfcfcf9,0xfbfbf6,
0xfafbf2,0xf9f9ef,0xebf2ef,0xcce5e9,0xa4d3dc,0xabd6da,0xdfebe2,0xd8eae3,
0xb5d0d3,0x97b0bb,0x7d97a8,0x69879c,0x587b92,0x4a718b,0x426f89,0x377084,
0x2f6e7d,0x2a717b,0x27767c,0x20787c,0x1f797c,0x1c7a7c,0x157b83,0x157e80,
0x087c46,0x00792d,0x007731,0x00742f,0x007731,0x00a244,0x00b14a,0x00ad49,
0x00ad4a,0x00ad49,0x00ad4a,0x00af4b,0x00ad4b,0x00ac4c,0x00ad4a,0x00ac4c,
0x00b047,0x089266,0x11648f,0x12614e,0x16551c,0x175627,0x175527,0x1a5126,
0x195527,0x059844,0x00b251,0x00aa51,0x00aa50,0x00ab4f,0x00ab4e,0x00ac4f,
0x00aa51,0x00aa50,0x00aa4e,0x00a84e,0x00a247,0x63c259,0xf5ee77,0xf4e976,
0xe7e170,0xdbd76b,0x76746e,0x231372,0x33197a,0x391f80,0x45348b,0x574e9b,
0x6565a8,0x7a80bc,0x92a4d1,0xb1ceec,0xd3ecfd,0xedf5fd,0xc3e6f8,0x7bcaf2,
0xa2d6f5,0xd6ebfb,0xf7fbfe,0xffffff,0xffffff,0x000000,0xffffff,0xfffffe,
0xfdfdf9,0xf9faf3,0xf6f6e9,0xf6f7e8,0xf7f8ef,0xe4efec,0xc1dfe5,0x90cddf,
0xaedae6,0xe5f1f1,0xcee8f3,0xa6c2de,0x8398c5,0x6578ae,0x5064a1,0x405596,
0x344a8e,0x2f4b8c,0x264e86,0x1a4e7f,0x0e5380,0x005783,0x005a85,0x006085,
0x006485,0x006a8d,0x00756a,0x007c36,0x007b33,0x007934,0x007531,0x008b3b,
0x00b04a,0x00ad48,0x00ad49,0x00ad4a,0x00ad49,0x00ad4a,0x00ae4a,0x02a949,
0x00ae4b,0x00ac4c,0x00ac4a,0x00ad4b,0x00ac4c,0x00ae4a,0x01a356,0x116066,
0x12542d,0x155b23,0x175927,0x175826,0x174f23,0x0d7333,0x00ad4f,0x00ae4f,
0x00ab50,0x00aa51,0x00ab4f,0x00ab4e,0x00ad4f,0x01a94e,0x01a84e,0x00ab51,
0x00a950,0x00a94e,0x00a84e,0x00a84d,0x00a448,0x18ab4c,0xcee370,0xffef79,
0xe8e272,0xe4df6d,0x9b976b,0x2d1e6e,0x351d7b,0x402583,0x453189,0x504594,
0x5f5ba2,0x7175b4,0x8896c9,0xa6bfe3,0xc9e7fa,0xe6f3fe,0xddeffb,0x8dd2f4,
0x95d2f5,0xcce7fa,0xf3f9fe,0xffffff,0xffffff,0x000000,0xffffff,0xfffffd,
0xfdfdfb,0xfcfcf8,0xfafaf3,0xfdfdf8,0xfbfcfd,0xe2f0fa,0xaedbf7,0x7ecbf4,
0xc0e5fa,0xeaf6ff,0xc3e1f8,0x9db4dc,0x7c8ac2,0x626bac,0x4f589d,0x404b93,
0x37468d,0x324b8c,0x2b4b83,0x234d7d,0x1c517f,0x436e7a,0x67887a,0x658d7d,
0x6d9784,0x56947c,0x0c8144,0x007f30,0x007f37,0x007a37,0x007e37,0x00a545,
0x00b14b,0x00ad48,0x00ad49,0x00ad4a,0x00ad49,0x00ad4a,0x00b04b,0x03a447,
0x106731,0x04a146,0x00b14e,0x00ac4b,0x00ad4b,0x00ac4c,0x00ad4c,0x00b24d,
0x078c3d,0x105923,0x115b26,0x135b27,0x165625,0x135c27,0x049b46,0x00b251,
0x00ab4e,0x00ab51,0x00ab4f,0x00ac4e,0x00b251,0x0d773a,0x09853f,0x00af52,
0x00ab51,0x00a94e,0x00a84e,0x00a84d,0x00a84a,0x00a147,0x74c65f,0xf8ec77,
0xeee573,0xe6e36e,0xb1ac69,0x39286a,0x2f1775,0x3f2481,0x412986,0x4a3b90,
0x57509c,0x696aac,0x7e89c1,0x9bb1da,0xbfdcf4,0xdef2fd,0xe9f4fc,0xa1d9f6,
0x8ccff5,0xc1e3f9,0xedf6fe,0xffffff,0xffffff,0x000000,0xffffff,0xfafcff,
0xdceefc,0xa4d8f7,0x82cdf4,0xcdeafa,0xe5f3fe,0xb9d7f2,0x95a8d4,0x7781ba,
0x6068a7,0x515b9c,0x465796,0x415b94,0x3b5f91,0x316187,0x296584,0x246b86,
0xa3b373,0xf3e66d,0xf5eb74,0xfffa7c,0x9dc75e,0x027f36,0x008037,0x008136,
0x007c36,0x00943f,0x00b04a,0x00ad49,0x00ad48,0x00ad4a,0x00ad49,0x00b64e,
0x0b803c,0x1c2c1e,0x0b833a,0x00b64f,0x00ac4a,0x00ad4a,0x00ac4c,0x00b14d,
0x01a647,0x0a6b2b,0x0f5923,0x115d26,0x125522,0x087f37,0x00b050,0x00af4f,
0x00ab4e,0x00ab50,0x00ab51,0x00aa4f,0x00b451,0x059b49,0x1b3b25,0x115f2e,
0x00ae51,0x00ad51,0x00a94e,0x00a84e,0x00a84d,0x00a84b,0x00a349,0x1aaa4e,
0xccdf6f,0xf5e875,0xe8e36e,0xbdb86a,0x423268,0x2a1470,0x3e2480,0x402684,
0x46338a,0x524897,0x6261a6,0x777eba,0x93a5d2,0xb5d3ee,0xd5eefd,0xecf5fd,
0xabddf6,0x86cdf3,0xb9e0f9,0xe8f3fd,0xfffeff,0xffffff,0xffffff,0x000000,
0xffffff,0xf6fbfe,0xd6ecfb,0x9bd4f6,0x84cdf3,0xd3edfa,0xdff1fd,0xb2cdec,
0x8d9ccd,0x7076b4,0x5b5fa3,0x4d5398,0x445093,0x415592,0x3a588b,0x305681,
0x2a5880,0x306382,0xb8be71,0xebe46f,0xf9ee77,0xf2ef76,0x51a74c,0x007c35,
0x00843b,0x008138,0x008239,0x00a545,0x00b04a,0x00ad49,0x00ad48,0x00ad49,
0x00ad4a,0x00ad49,0x00af4a,0x00af4b,0x115a2f,0x1d1f1b,0x135f2d,0x01af4b,
0x00b14d,0x00ad4a,0x00ac4c,0x00af4c,0x00b14d,0x028637,0x075b20,0x0d5a22,
0x0d6227,0x029c46,0x00b351,0x00ad4e,0x00ac4f,0x00ab51,0x00ad4f,0x00b652,
0x0d753c,0x21241e,0x164b25,0x039e48,0x00af51,0x00a94e,0x00a84e,0x00a84d,
0x00a84b,0x00a74a,0x00a24a,0x82c960,0xf8e976,0xe9e46f,0xc9c46a,0x504169,
0x25106c,0x3b227c,0x402484,0x432d88,0x4d4092,0x5d58a1,0x7075b4,0x8b99cb,
0xadc8e8,0xcfe9fa,0xebf6fd,0xb4e1f6,0x83ccf3,0xb2ddf7,0xe3f1fc,0xfdfefe,
0xffffff,0xffffff,0x000000,0xffffff,0xf3f9fe,0xd2e9fa,0x94d2f5,0x87cff3,
0xd9eefa,0xd7edfc,0xa9c3e5,0x8693c7,0x6d6eb0,0x58529d,0x494090,0x40378a,
0x3d3888,0x35357a,0x313171,0x2b2e73,0x3b4073,0xc3bf6e,0xebe471,0xffef79,
0xaacd64,0x0f883e,0x00853b,0x00873c,0x008139,0x00903e,0x00ae49,0x00ad49,
0x00ad48,0x00ad49,0x00ad4a,0x00ad48,0x00b24b,0x039640,0x124c24,0x1b3d20,
0x174f25,0x079440,0x00b64f,0x00ad4b,0x00ac4c,0x00ae4b,0x00b24e,0x009f45,
0x016625,0x05581d,0x057c32,0x01ae4f,0x00af50,0x00ac4e,0x00ab50,0x00b051,
0x00a54b,0x115f2d,0x1b3921,0x164923,0x097c37,0x00ae51,0x00aa4e,0x00a84e,
0x00a74d,0x00a84a,0x00a84b,0x00a24a,0x2fb253,0xe4e372,0xefe570,0xd3cf6b,
0x615368,0x210c67,0x382175,0x3f2382,0x412986,0x49398f,0x59519d,0x6b6daf,
0x8490c5,0xa5bde1,0xc7e3f8,0xebf6fd,0xc1e5f8,0x84ccf3,0xabdaf7,0xdfeffc,
0xfbfdfe,0xffffff,0xffffff,0x000000,0xffffff,0xf1f7fe,0xcce7fa,0x8fd0f4,
0x90d3f4,0xdef1fc,0xd0e8f9,0xa1b9df,0x8089c1,0x6767ab,0x544b98,0x46388b,
0x3f2d86,0x3a2b80,0x332872,0x30296c,0x2a2971,0x444773,0xcbc66c,0xf5e973,
0xe8e674,0x47a44d,0x00823b,0x008b40,0x00883e,0x00863b,0x00a145,0x00af49,
0x00ad49,0x00ad48,0x00ad49,0x00ad4a,0x00ab48,0x0db54d,0x76ba56,0x2d642d,
0x105724,0x185323,0x106d2f,0x02a949,0x00b14d,0x00ac4c,0x00ad4b,0x00af4b,
0x00af4d,0x007f34,0x006123,0x009b44,0x00b152,0x00ac4e,0x00ab4e,0x00ab50,
0x00ac50,0x05b050,0x10883c,0x105724,0x125c26,0x135825,0x10602a,0x039e49,
0x00ac50,0x00a84e,0x00a74d,0x00a74a,0x00a64b,0x00a64c,0x03a44b,0x9ecf65,
0xfae873,0xdbd76b,0x716469,0x210b67,0x351e70,0x3e2380,0x402684,0x47348b,
0x544a99,0x6665aa,0x7d87c0,0x9db3db,0xbfdcf4,0xe7f5fe,0xcbeafa,0x85cdf3,
0xa5d8f6,0xdbedfc,0xf9fcfe,0xffffff,0xffffff,0x000000,0xffffff,0xeef7fe,
0xc7e5f9,0x8ccff4,0x9ad7f6,0xe2f3fe,0xc9e2f6,0x9bb0da,0x7b82bc,0x6260a6,
0x504495,0x433189,0x402782,0x3a2478,0x32206b,0x32256a,0x2a2a72,0x495375,
0xd1cc6e,0xfbeb75,0xa4ca63,0x058b3f,0x008b41,0x008c40,0x00873f,0x009642,
0x00ae49,0x00ae48,0x00ad49,0x00ad48,0x00ad49,0x00ad4a,0x00a845,0x4dc256,
0xf3f87e,0x779849,0x054e20,0x185625,0x165323,0x088a3b,0x00b54f,0x00ad4c,
0x00ac4c,0x00ad4b,0x00ad4a,0x00b14e,0x009943,0x008639,0x00ab4e,0x00ae50,
0x00ab4e,0x00ac4e,0x00ac4f,0x00ab51,0x00a74e,0x45c35b,0x88af53,0x195826,
0x095724,0x135c27,0x135524,0x09843b,0x00ae51,0x00a84e,0x00a84d,0x00a74a,
0x00a64b,0x00a64d,0x00a24b,0x3bb354,0xe5e370,0xe7de6d,0x7d7068,0x210c65,
0x321b6b,0x3d217d,0x3f2584,0x432f88,0x504595,0x615fa6,0x787fba,0x97aad6,
0xb8d6f1,0xe3f4fe,0xd0ecfa,0x85cdf3,0xa1d6f5,0xd7ecfb,0xf8fbfe,0xffffff,
0xffffff,0x000000,0xffffff,0xeef7fd,0xc6e5fa,0x8bcff4,0x9ed9f6,0xe3f4fe,
0xc4ddf4,0x96abd5,0x767db8,0x5f5ba3,0x4d3f92,0x432c87,0x3f2481,0x381f74,
0x301b67,0x321d66,0x281c6a,0x47466e,0xd8cf70,0xe9e571,0x47a850,0x008a3e,
0x009142,0x008c41,0x00a647,0x00b049,0x00ad48,0x00ad49,0x00ad48,0x00ad49,
0x00ab49,0x05aa46,0xa2da66,0xffff82,0xc5d166,0x195926,0x125524,0x185123,
0x106d2f,0x02ab4b,0x00b14e,0x00ab4c,0x00ac4c,0x00ad4b,0x00ad4a,0x00ad4c,
0x00ab4d,0x00aa4c,0x00ab4e,0x00ac4e,0x00ab4f,0x00a94f,0x09a84d,0x9fe06c,
0xf4ef77,0x3f7636,0x004d20,0x145c27,0x145624,0x0f692e,0x01a64b,0x00aa4f,
0x00a84e,0x00a84d,0x00a84a,0x00a74b,0x00a64d,0x00a44d,0x06a34a,0xa0cf63,
0xf2e26f,0x82796a,0x241268,0x301b6a,0x392179,0x3f2483,0x422b87,0x4e4093,
0x5e5aa3,0x7479b6,0x92a4d1,0xb3d0ee,0xe0f2fd,0xd2ecfa,0x84cdf3,0x9ed5f6,
0xd6ecfb,0xf7fbfe,0xffffff,0xffffff,0x000000,0xffffff,0xedf6ff,0xc4e4fc,
0x88cdf6,0xa1dbf8,0xe4f5ff,0xc0d9f3,0x92a5d3,0x7177b5,0x5a55a0,0x48398e,
0x3e2685,0x3b207f,0x331b70,0x2e1665,0x2f1764,0x261166,0x463869,0xdecf70,
0x9fcb61,0x049144,0x009143,0x009242,0x008d41,0x009944,0x00ae48,0x00af48,
0x00ad48,0x00ad49,0x00ad48,0x00ad49,0x00a745,0x42bd52,0xe7f177,0xfffc7e,
0xf6f67a,0x52813b,0x064d20,0x195525,0x155a26,0x059842,0x00b34f,0x00ac4c,
0x00ad4a,0x00ad4b,0x00ac4c,0x00ad4e,0x00ad4f,0x00ab4e,0x00ac4e,0x00ac4f,
0x00ab50,0x00a54e,0x32b755,0xeff579,0xffff82,0x84a74f,0x055122,0x0d5725,
0x135927,0x125926,0x059443,0x00ae51,0x00a84e,0x00a84c,0x00a84a,0x00a64b,
0x00a54d,0x00a14a,0x4cb955,0xeae06e,0x8a846e,0x241d6c,0x2d1f6a,0x362275,
0x3d2482,0x402985,0x4a3a8f,0x5a539f,0x7073b3,0x8e9ecd,0xafcceb,0xddf0fd,
0xd4edfb,0x83cef2,0x9dd5f6,0xd6ebfb,0xf6fafe,0xffffff,0xffffff,0x000000,
0xffffff,0xfefefc,0xfbfbf4,0xeaf2f2,0xc6e2ef,0x92cfea,0xa8dbed,0xe2f2f5,
0xbfd7e9,0x95a6cc,0x797db2,0x635ea0,0x524590,0x4b3488,0x462f82,0x3d2972,
0x39246a,0x39256a,0x2e1d6a,0x53436d,0xcdca6c,0x42ac4e,0x009042,0x009445,
0x009244,0x009143,0x00a547,0x00b048,0x00ad48,0x00ad49,0x00ad48,0x00ab48,
0x00a945,0x98d764,0xfff97d,0xfef67a,0xfffe7e,0xb6c861,0x195b27,0x115524,
0x195223,0x0d7b35,0x01b04e,0x00af4e,0x00ab4c,0x00ad4a,0x00ad4b,0x00ac4c,
0x00ac4d,0x00ab4e,0x00ac4e,0x00a94e,0x0ba94f,0x9ad868,0xfffa7c,0xffff82,
0xbbcc64,0x195e28,0x055222,0x145b27,0x135625,0x0a7835,0x00ab4f,0x00a94f,
0x00a84c,0x00a84a,0x00a64b,0x00a64d,0x00a54d,0x00a24b,0x1fac4e,0xccd969,
0x94856d,0x1c1567,0x281a65,0x331d70,0x3a207d,0x3d2382,0x45358d,0x564f9c,
0x6c6faf,0x8b99cb,0xacc8e8,0xdcf0fc,0xd4eefb,0x83cdf2,0x9dd5f6,0xd6ebfb,
0xf6fafe,0xffffff,0xffffff,0x000000,0xffffff,0xfefefd,0xfbfaf5,0xebf2f3,
0xc7e4ef,0x95d1ea,0xa6d9eb,0xdff0f3,0xc0d6e6,0x98a7ca,0x7f81b2,0x6c65a1,
0x5e4f95,0x58438e,0x554189,0x4d3b7a,0x4a3975,0x4c3a76,0x433478,0x695379,
0x8fb661,0x0a9c46,0x009645,0x009644,0x009145,0x009a46,0x00ad48,0x00b048,
0x00ad48,0x00ad49,0x00ae49,0x00a746,0x32b750,0xdaec73,0xfff97c,0xfbf479,
0xfffa7c,0xf3f57a,0x5f8a40,0x044c1f,0x185524,0x145d28,0x069943,0x00b24f,
0x00ac4c,0x00ad4b,0x00ac4c,0x00ac4d,0x00ac4e,0x00ab4e,0x00ac4e,0x00ac4f,
0x00a74d,0x45bd59,0xf1f478,0xfff67a,0xfffa7c,0xf4f279,0x5a8940,0x025021,
0x115a26,0x135826,0x116029,0x039b47,0x00ad50,0x00a84b,0x00a84a,0x00a64c,
0x00a64d,0x00a54d,0x00a44c,0x04a549,0x72c157,0xaea17f,0x5d5694,0x61578f,
0x695a95,0x705fa1,0x7361a5,0x796dab,0x877fb6,0x9696c5,0x9ca7d2,0xa7c3e4,
0xdbf0fb,0xd2edfa,0x85cdf3,0x9fd6f6,0xd7ecfb,0xf7fbfe,0xffffff,0xffffff,
0x000000,0xffffff,0xeff8ff,0xc8e6fa,0x8ecff3,0x9cd7f3,0xddf1fa,0xbed5ec,
0x909fcc,0x7276b1,0x5d589d,0x4e418f,0x483388,0x453181,0x3c2c70,0x3a296c,
0x3c2c6d,0x3b2770,0x4b3c6f,0x2b9951,0x009d46,0x009948,0x009645,0x009746,
0x00a848,0x00b048,0x00ad48,0x00ad49,0x00ac48,0x01a846,0x88d262,0xfff97c,
0xfff67a,0xfbf479,0xfdf67a,0xffff7f,0xbccc64,0x165825,0x0d5323,0x185123,
0x0e7332,0x01ad4d,0x00b04e,0x00ad4b,0x00ad4c,0x00ac4c,0x00ac4e,0x00ab4e,
0x00ac4e,0x00ab4d,0x07a94e,0xa7dc6b,0xfffa7c,0xfbf479,0xfbf378,0xffff82,
0xafc561,0x125a27,0x095524,0x135a27,0x135725,0x097f39,0x00af51,0x00a84a,
0x00a849,0x00a64b,0x00a64d,0x00a54d,0x00a449,0x24b04c,0x778c72,0x494085,
0x423c7b,0x483b7f,0x503f8d,0x534092,0x584a97,0x655fa4,0x7879b5,0x8c9ac9,
0xa6c1e4,0xdaeefa,0xcfebfa,0x85cdf3,0xa1d6f5,0xd8ecfb,0xf8fbfe,0xffffff,
0xffffff,0x000000,0xffffff,0xf0f8fe,0xcde7fa,0x91d0f2,0x94d4f1,0xdaeff8,
0xc0d6ea,0x90a1ca,0x727aaf,0x5f609d,0x525091,0x4d4b8d,0x4a4b86,0x434679,
0x414677,0x404a77,0x47487d,0x34646e,0x019c4b,0x00a048,0x009a48,0x009747,
0x009e45,0x00af47,0x00b048,0x00ae48,0x00ad48,0x00ad49,0x00a844,0x23b34d,
0xcee971,0xfffa7d,0xfcf479,0xfbf479,0xfffb7d,0xeef278,0x50803b,0x004b1d,
0x185726,0x145b27,0x059b44,0x00b24d,0x00ad4a,0x00ad4b,0x00ac4d,0x00ac4e,
0x00ab4e,0x00ac4e,0x00ad4e,0x00a64b,0x2fb654,0xecf077,0xfff67b,0xfbf479,
0xfbf379,0xfffb7c,0xedee76,0x3f7736,0x025122,0x125b27,0x145624,0x0e682d,
0x01a44d,0x00ab4d,0x00a749,0x00a64b,0x00a54d,0x00a64b,0x06ab49,0x277661,
0x2e2b77,0x27296e,0x2d2670,0x37277e,0x3b2a84,0x41358a,0x504c98,0x646aac,
0x8191c5,0xa4bfe1,0xd7eaf9,0xc7e5f8,0x82caf0,0xa3d7f4,0xdaedfb,0xf9fcfe,
0xffffff,0xffffff,0x000000,0xffffff,0xf3f9fe,0xd2eafb,0x94d1f6,0x88d1f5,
0xd5f0ff,0xc2d8f3,0x899cce,0x666faf,0x4e549b,0x3d4590,0x34418b,0x2f4285,
0x28417a,0x2b4a7a,0x32547e,0x2d577f,0x0d8a5c,0x00a448,0x009f4b,0x009c4a,
0x009a48,0x00a747,0x00b048,0x00ae48,0x00ad48,0x00ad49,0x00ac47,0x00a744,
0x69ca5c,0xfaf87c,0xfef67a,0xfbf479,0xfef57a,0xffff7f,0xaabe5d,0x125624,
0x0e5424,0x185424,0x0c7b36,0x00b14d,0x00af4c,0x00ac4b,0x00ac4d,0x00ac4e,
0x00ab4e,0x00ac4e,0x00ac4d,0x04a84c,0x96d567,0xfff97c,0xfcf479,0xfbf479,
0xfdf479,0xffff7f,0x88aa51,0x085423,0x0d5825,0x135926,0x125925,0x07873c,
0x00af4e,0x00a74a,0x00a64c,0x00a64d,0x00a54d,0x00a64c,0x00ab48,0x078d57,
0x27487a,0x2d3c7b,0x2d3f7b,0x334287,0x37448c,0x3b4c90,0x485b9a,0x5b72a9,
0x7592be,0x91b8d3,0xbfdae4,0xa9d5e5,0x74c3e4,0x9dd1ea,0xd1e7f2,0xeff5f7,
0xfafbfa,0xfefefe,0xffffff,0xffffff,0x000000,0xffffff,0xfffffe,0xfffefe,
0xf6fafd,0xd6ecfa,0x9cd4f5,0x87cef1,0xd0edfa,0xc5dcf0,0x8ca1ce,0x6775af,
0x4d5a9c,0x3a4d91,0x304c8d,0x2a4e88,0x204e7f,0x285f7f,0x336687,0x227d74,
0x02a14b,0x00a148,0x009d4b,0x00a249,0x00ad48,0x00ae48,0x00ad48,0x00ad49,
0x00a945,0x12af49,0xb8e36c,0xfffc7e,0xfbf479,0xfbf379,0xfaf279,0xfffb7d,
0xf3f57a,0x5c8840,0x00491d,0x165625,0x145e28,0x04a045,0x00b24d,0x00ad4b,
0x00ac4d,0x00ab4e,0x00ac4e,0x00a64b,0x3fba57,0xeef279,0xfff67b,0xfaf379,
0xfbf379,0xfbf479,0xfaf379,0xfffd7e,0xd6de6e,0x24662d,0x055321,0x135b27,
0x145624,0x0d6d2e,0x00a649,0x00ab4c,0x00a54c,0x00a54d,0x00a54c,0x00a749,
0x02a44b,0x107a6e,0x196083,0x146682,0x126a89,0x136b8e,0x136d91,0x1c7295,
0x307c9d,0x478fa8,0x5ca5b4,0x86bebf,0x77bfc6,0x64becf,0x90ccd7,0xbddce0,
0xdce7e9,0xf0f4f5,0xfefdfe,0xffffff,0xffffff,0x000000,0xffffff,0xfffefe,
0xf9fcfe,0xdceffa,0xa5d8f5,0x86cdf0,0xcbeaf7,0xcce1f0,0x94a9d0,0x6d79b1,
0x4e589d,0x3b4790,0x31458b,0x2b4988,0x224a7c,0x1f4f7d,0x1c5a82,0x098f5c,
0x00a647,0x00a04a,0x009d4a,0x009e4b,0x00aa49,0x00af48,0x00ae48,0x00ad48,
0x00ac49,0x00a745,0x56c356,0xf4f579,0xfff77b,0xfbf479,0xfcf579,0xfff977,
0xfffa78,0xffff80,0xb8c95f,0x185925,0x0a5222,0x185324,0x0d7733,0x00b04b,
0x00af4d,0x00ac4d,0x00ab4e,0x00ac4e,0x00aa4d,0x0fac4c,0xb2e169,0xffff7c,
0xfff979,0xfff978,0xfff877,0xfffa78,0xfffd7c,0x82a54b,0x095422,0x105b27,
0x125925,0x125c27,0x04903f,0x00af4e,0x00a64c,0x00a54d,0x00a54c,0x00a34b,
0x00a648,0x009d58,0x00877a,0x008281,0x008383,0x008086,0x007f89,0x007f8d,
0x048692,0x169298,0x40a2a3,0x74b9b2,0x68bdbc,0x64bfc6,0x91ced1,0xb9dbdb,
0xd9e6e6,0xeef2f3,0xfbfafc,0xffffff,0xffffff,0x000000,0xffffff,0xfdfeff,
0xe3f1fd,0xb0dcf8,0x80cbf3,0xbee6f9,0xd4e8f7,0x98aed8,0x7980b4,0x6d6d9d,
0x5a5691,0x4c498b,0x434885,0x3b497b,0x3a447e,0x1f676d,0x01a04a,0x00a448,
0x00a04a,0x009f4a,0x00a34a,0x00ae48,0x00ad48,0x00aa46,0x12ae49,0xb1df6a,
0xfffa7e,0xfcf479,0xfbf479,0xfcf478,0xf9f479,0xedf07a,0xe9f07b,0xe9f07a,
0xeef580,0xd8e576,0x417837,0x044d1f,0x165626,0x145a27,0x059b42,0x00b14e,
0x00ac4d,0x00ac4e,0x00ab4e,0x00ac4e,0x00a74a,0x48bf58,0xe0ef79,0xebf37b,
0xe8ef7a,0xe9f079,0xeaf07a,0xebf17a,0xecf079,0xebef7a,0xedf680,0xc8d76e,
0x1f632a,0x0d5825,0x125c27,0x135623,0x0b7632,0x00aa4c,0x00aa4e,0x00a54d,
0x00a54c,0x00a34b,0x00a449,0x00a549,0x02966a,0x058a7d,0x06897d,0x068780,
0x048583,0x038787,0x058e8d,0x209a97,0x50aba5,0x78bcb2,0x56b5b9,0x57b4c2,
0x8ec7cf,0xc2dfde,0xe2edeb,0xf0f5f4,0xfbfcfc,0xffffff,0xffffff,0x000000,
0xffffff,0xfffffe,0xfefefb,0xfefdf8,0xfefdfa,0xe9f4fc,0xbee1fa,0x7ecbf2,
0xa3dcf8,0xd6ebfd,0xa1b7e0,0x7d83b9,0x726ea1,0x625491,0x584388,0x514181,
0x513e77,0x514b76,0x1e8b5a,0x00a748,0x00a14b,0x00a04b,0x00a04a,0x00a849,
0x00af48,0x00ae48,0x00ad48,0x00ac48,0x00a845,0x4cc255,0xf9f67b,0xfef67b,
0xfbf479,0xfff878,0xe9eb7c,0x62a492,0x429491,0x519a91,0x479e98,0x4a9a97,
0x2c775c,0x125620,0x145824,0x165424,0x0c7f36,0x00b34e,0x00ac4d,0x00ac4e,
0x00ab4e,0x00ad4d,0x04ac4b,0x36a76f,0x55a682,0x51a67b,0x4da37d,0x4a9f80,
0x4a9f85,0x51a888,0x5cb18d,0x55b094,0x45a597,0x439f98,0x1a6e4c,0x115720,
0x115c27,0x135624,0x106228,0x039c45,0x00ad50,0x00a54d,0x00a54c,0x00a44b,
0x00a34b,0x00a547,0x099c5b,0x148e7c,0x0f8c81,0x0a8984,0x078787,0x0a898b,
0x1b8f94,0x3f9ea1,0x6ab5b0,0x7ec0bc,0x4fb4c1,0x64b9cc,0xa0cfdb,0xd6e9eb,
0xf0f6f5,0xf8fafa,0xfdfefa,0xfcfcf7,0xfcfdf9,0xfffefe,0xffffff,0xffffff,
0x000000,0xffffff,0xfffffe,0xfdfdf9,0xfafaf4,0xfbfbf4,0xfbfbf2,0xedf4f4,
0xc9e5f5,0x8bcdf0,0x90d3ef,0xd3eaf6,0xb0c4e2,0x7d89bf,0x605da5,0x4d3e91,
0x422a86,0x3a207e,0x360d6f,0x254060,0x069e4d,0x00a74b,0x009f4b,0x009f4a,
0x00a34a,0x00ad48,0x00ae48,0x00ad48,0x00ab48,0x0aab47,0x9cda66,0xfffb7e,
0xfbf479,0xfff878,0xeced7c,0x378c99,0x00679b,0x14709c,0x0075a3,0x046c9f,
0x0e7088,0x126038,0x15581e,0x145625,0x12632a,0x02a448,0x00b050,0x00ab4e,
0x00b04a,0x04a05b,0x138883,0x0e8485,0x0c857e,0x088282,0x037c84,0x027a89,
0x08878c,0x139393,0x0c929c,0x00839f,0x0080a7,0x077274,0x125a22,0x135b25,
0x115a27,0x135623,0x088038,0x00aa4f,0x00a84e,0x00a54d,0x00a54c,0x00a44b,
0x00a34b,0x00a548,0x02a24e,0x0e8b79,0x13828d,0x15828d,0x1d8390,0x2a8597,
0x4390a3,0x65a9b5,0x93c9c8,0x8ecccf,0x62bdd2,0x8ccbd9,0xbfdee2,0xe2ede8,
0xedf2ec,0xeef4ed,0xf1f5e7,0xf6f7e9,0xfcfcf7,0xffffff,0xffffff,0x000000,
0xffffff,0xfefefc,0xfafbf5,0xfbfbf6,0xfbfbf3,0xf0f5f0,0xd2e9f1,0x9cd3ef,
0x89ceeb,0xcce9f0,0xbed2e3,0x8d9bc5,0x7373ae,0x5e539b,0x544290,0x5c3f8e,
0x533a7f,0x177858,0x00a848,0x00a14c,0x00a04b,0x00a948,0x00af47,0x00ae48,
0x00ad48,0x00a946,0x37ba51,0xeaf076,0xfff77b,0xfbf479,0xfff678,0xfbf779,
0x549892,0x055b98,0x1a6499,0x016ca1,0x095b93,0x06629e,0x096469,0x155a22,
0x165923,0x155625,0x098439,0x00b250,0x00ab4e,0x00ae4c,0x00ad4d,0x028474,
0x02728e,0x02798a,0x027a86,0x017c89,0x027a89,0x037a8c,0x0d888e,0x199395,
0x11929e,0x03879f,0x0285a4,0x097d96,0x126039,0x125a20,0x125d27,0x135623,
0x0e682c,0x029e49,0x00aa50,0x00a54d,0x00a54c,0x00a44b,0x00a34b,0x00a44b,
0x00a645,0x208f6e,0x457a99,0x4b7e99,0x58889c,0x6a93a5,0x80a6b3,0x9ec4c5,
0xc1dfd8,0xb1dadd,0x98d0d9,0xb9dbda,0xd5e6db,0xe7eddf,0xeaf0e1,0xedf0e0,
0xf7f8ed,0xfefefc,0xfffffe,0xffffff,0xffffff,0x000000,0xffffff,0xfefdfb,
0xf4f6ee,0xdaebeb,0xb1d9e8,0x8dcce3,0xbee1e5,0xcbdddf,0xa1afc6,0x898db4,
0x736da4,0x68589b,0x75549f,0x547684,0x09a64d,0x00a449,0x009f4c,0x00a04c,
0x00a44a,0x00ae46,0x00af48,0x00ae48,0x00ad48,0x00ac47,0x03a946,0x95d664,
0xfff97d,0xfcf479,0xfbf479,0xfef579,0xfffa78,0x76ab8e,0x0f639a,0x296c9c,
0x0e75a5,0x196494,0x16699d,0x0f7299,0x155f39,0x15581b,0x155624,0x12632c,
0x03a248,0x00b050,0x00af48,0x129e5f,0x2e7f87,0x2d828f,0x2c878c,0x2d878b,
0x2c8b8f,0x32888f,0x388791,0x489493,0x549e9b,0x4a9da6,0x3491a8,0x348ca9,
0x3f86aa,0x2e6b61,0x0f591f,0x125c27,0x125a25,0x115a24,0x068c3e,0x00ab50,
0x00a64d,0x00a54c,0x00a44b,0x00a34b,0x00a34c,0x00a444,0x189f56,0x79989f,
0x99a0b4,0x9aa6b3,0x9daaba,0xa6b6c6,0xbfd3d6,0xd7e8de,0xbadad5,0xb0d5d0,
0xcce0d3,0xe1ebda,0xf4f5e7,0xfafaee,0xfcfbf4,0xfffefc,0xffffff,0xffffff,
0x000000,0xffffff,0xfffffe,0xfdfdfb,0xe6f1f7,0xbce0f4,0x8accea,0xa8dae8,
0xd1e6eb,0xa7b9d0,0x868db3,0x7573a0,0x6c5997,0x5c4f8c,0x1e8a60,0x00ac47,
0x00a14c,0x00a04b,0x009f4c,0x00a948,0x00af47,0x00ae48,0x00ad48,0x00a945,
0x30b84f,0xe0ee74,0xfff77b,0xfbf479,0xfcf479,0xfffc77,0x92bc8b,0x18699b,
0x2c719e,0x1179a7,0x216896,0x206b9a,0x1476ad,0x146968,0x125921,0x13561c,
0x135620,0x09833b,0x00b54e,0x01ac4e,0x28847b,0x487193,0x447a92,0x447a91,
0x487a90,0x477e94,0x497b94,0x4c7b96,0x568a98,0x6196a1,0x5696aa,0x448aab,
0x4487ac,0x4e84ae,0x49768d,0x1a5d2f,0x0f5a22,0x115c29,0x135623,0x0b7532,
0x00a64d,0x00a84e,0x00a54c,0x00a44b,0x00a34b,0x00a449,0x01a746,0x499079,
0x707999,0x68799b,0x7083a5,0x8aa0ba,0xbdd1d7,0xcbe5e3,0x8fcee1,0x9bd2e7,
0xcfe6ed,0xeef4f1,0xfefdf9,0xfffefe,0xffffff,0xffffff,0x000000,0xffffff,
0xf2f9ff,0xcae8fb,0x90d0f6,0x86cff4,0xceebfc,0xb0c2e4,0x7c87bf,0x6662a4,
0x5f4299,0x3e577c,0x06a24d,0x00a84a,0x00a14d,0x009f4b,0x00a24b,0x00ae45,
0x00ae47,0x00ae48,0x00ad48,0x00a945,0x8dd663,0xfffa7d,0xfcf479,0xfbf479,
0xfcf479,0xfffc77,0xb3cf84,0x226c97,0x2d6f9b,0x1476a3,0x226593,0x1f6894,
0x1173a4,0x15739d,0x196870,0x1b603f,0x165721,0x0e6723,0x00ac40,0x11a262,
0x31758a,0x33768c,0x317d8b,0x347d8a,0x377d88,0x35808d,0x367c8c,0x3e7c8c,
0x4e8c90,0x569597,0x4b939e,0x3a89a0,0x3885a2,0x4084a3,0x4c7e9a,0x2c6b4f,
0x0d591d,0x125d29,0x125825,0x105e27,0x039845,0x00ab4f,0x00a54c,0x00a44b,
0x00a34b,0x00a24b,0x00a846,0x0a8b5a,0x3b508f,0x4f5d9e,0x667dad,0x8eaacb,
0xcae2ee,0xb8e4f6,0x7dcbf3,0xa7d9f7,0xdceffc,0xfafdff,0xffffff,0xffffff,
0x000000,0xffffff,0xf8fdfe,0xd8edfb,0xa8d8f6,0x79c8f2,0xc0e7fb,0xc2d6ef,
0x8797c9,0x6761ad,0x4e3f95,0x187762,0x00ae45,0x00a24d,0x00a14c,0x00a04c,
0x00a948,0x00af44,0x00ae48,0x00ad48,0x00ab47,0x16b14b,0xd8eb72,0xfff77b,
0xfbf479,0xfffa78,0xdfe67e,0x3a7092,0x276198,0x146ea1,0x255a8f,0x265d90,
0x116b9f,0x146ba1,0x2168a1,0x21668f,0x1c5d6a,0x16583f,0x088241,0x1f8377,
0x37668e,0x336e89,0x31728c,0x32728c,0x33728a,0x2c748e,0x2a6f8c,0x30718e,
0x428490,0x458a98,0x31859e,0x1f7b9e,0x1a77a0,0x1978a1,0x23709e,0x1f6871,
0x105e20,0x0f5e1c,0x105c22,0x135723,0x087f39,0x00a94e,0x00a74d,0x00a44b,
0x00a34b,0x00a34a,0x00a24b,0x00a549,0x02a24b,0x356882,0x5e5eaa,0x7385bc,
0xa1bbde,0xd5edfb,0x9ed9f5,0x81cbf2,0xbce1f9,0xe7f4fd,0xfeffff,0xffffff,
0xffffff,0x000000,0xffffff,0xfdfeff,0xe5f3fc,0xbde1f8,0x84ccf3,0x9ad8f6,
0xcbe5f7,0x9cafd8,0x7573b7,0x56569b,0x1b8562,0x00ae45,0x00ac48,0x00a14d,
0x00a14a,0x00ad45,0x00ae46,0x00ad48,0x00ae48,0x00a845,0x5dc557,0xfdf67a,
0xfef57a,0xfbf479,0xfff779,0xf8f679,0x58808b,0x115197,0x09669e,0x1b4d8a,
0x1e4f8b,0x04619e,0x03609c,0x145793,0x165a9a,0x1b5699,0x264a81,0x214e76,
0x215082,0x234d87,0x204f85,0x1d5288,0x1d5287,0x1d5186,0x185489,0x175187,
0x1a5287,0x2b648c,0x306e94,0x186999,0x086097,0x076299,0x02649d,0x105695,
0x1d4c82,0x1f4e51,0x185835,0x115e26,0x0f5b1d,0x0d6324,0x029c46,0x00a84e,
0x00a34b,0x00a24b,0x00a24a,0x00a34a,0x00a24c,0x00a746,0x189161,0x6572af,
0x8498c9,0xb5d1e5,0xc4e8f5,0x81ccf0,0x97d2f5,0xcee8fb,0xf2f9fe,0xffffff,
0xffffff,0x000000,0xffffff,0xf2f9fe,0xcee8fa,0x9fd5f6,0x79caf3,0xc2e8fa,
0xb5cbe9,0x7f8ac1,0x635ea7,0x48478d,0x275b6e,0x0c9055,0x00ab48,0x00b544,
0x00ba41,0x00b842,0x00b744,0x00b844,0x00b245,0x00ab46,0x00aa45,0x00a743,
0x06a945,0xb7e06a,0xfff87b,0xfcf479,0xfbf479,0xfdf579,0xfdfc78,0x9eb084,
0x185698,0x07639e,0x1c4b88,0x204c89,0x095e9b,0x0a5e9a,0x195390,0x1c518f,
0x224a8b,0x2c3c81,0x2e3780,0x2b387c,0x293a7a,0x2b397b,0x293b7c,0x2a3a7b,
0x2a3a7a,0x293f7e,0x293c7b,0x2b3b7c,0x2e4480,0x38558a,0x2e5a91,0x1e4f8f,
0x175495,0x145999,0x244a8d,0x2d3a7e,0x2c377c,0x324779,0x2d5066,0x1b5146,
0x14552a,0x058636,0x00b14a,0x00ae48,0x00ad48,0x00ad47,0x00ac46,0x00ab47,
0x00aa44,0x00a641,0x4d8e95,0x9babda,0xc8e3f2,0xa2dcf5,0x77c8f0,0xb4ddf8,
0xddeffc,0xfbfdff,0xffffff,0xffffff,0x000000,0xffffff,0xfafdfe,0xdeeffc,
0xb8def9,0x7ecaf3,0xa9dff8,0xcce3f7,0x91a4d2,0x6d72b2,0x58449c,0x49238c,
0x392d7e,0x1e6664,0x118a50,0x118951,0x108b57,0x0e8a59,0x0e8758,0x2fa555,
0x52c152,0x4ebf53,0x49bb53,0x7bcb5d,0xf2f278,0xfff67a,0xfbf479,0xfffc78,
0xe8e47a,0x467999,0x0160a2,0x24518a,0x264f8a,0x125d9a,0x135c99,0x20508f,
0x234d8d,0x274486,0x2d3377,0x2e3174,0x2f3275,0x2e3074,0x2f3075,0x2f3074,
0x2e2f74,0x2e3378,0x2e2f75,0x302d74,0x2f3378,0x3b4a84,0x35518c,0x25478b,
0x1a5093,0x1d5497,0x284489,0x2d3278,0x2d2d74,0x3a3a7f,0x3e3a81,0x313275,
0x2f3367,0x224b58,0x137a59,0x137859,0x137359,0x16715a,0x1a725f,0x1d7163,
0x1f7667,0x24806d,0x298b6d,0x5a999c,0xb5c6e9,0xcdeafb,0x84cff4,0x8ed0f4,
0xc8e5fa,0xeaf5fd,0xffffff,0xffffff,0x000000,0xffffff,0xedf5fe,0xc9e6fa,
0x95d2f5,0x86cff3,0xd0ebfb,0xacc1e3,0x7c85be,0x5d5aa2,0x4a368e,0x431d86,
0x3c1f7e,0x33216d,0x2f2168,0x2c3075,0x2c3278,0x232275,0x5d5e72,0xd6d26d,
0xedea72,0xeae875,0xf7f079,0xfcf37b,0xfbf379,0xfbf479,0xfff878,0xfcf878,
0x79a997,0x0067a9,0x205d90,0x1f5a90,0x08649d,0x06619c,0x155692,0x1e5390,
0x254a89,0x2c3a7b,0x2c3879,0x2d377a,0x2d3579,0x2d3578,0x2d3477,0x2e3377,
0x2f3177,0x2c357a,0x2e2f76,0x2f2c73,0x313478,0x3d4985,0x354f8b,0x25498c,
0x195294,0x205295,0x2a4186,0x2f2e75,0x2f276d,0x3b3173,0x3b3174,0x31266f,
0x332772,0x342771,0x33296f,0x33276f,0x31236c,0x352371,0x3c257f,0x412586,
0x49338e,0x5a519f,0x7172b4,0x92a2d0,0xc7dff4,0xb0e1f9,0x7ac9f1,0xabdaf7,
0xd7ecfc,0xf6fbff,0xffffff,0xffffff,0x000000,0xffffff,0xf9fcff,0xdceefc,
0xb3ddf8,0x7bc9f2,0xa2dcf7,0xcadff5,0x8fa1d1,0x6a6dae,0x514795,0x432c87,
0x3f2083,0x381c77,0x31196a,0x2f2975,0x2f2d79,0x291b72,0x312373,0xa99d6f,
0xfbf071,0xf2e876,0xf8ef78,0xf7f17a,0xf9f37a,0xfaf378,0xfbf479,0xfcf578,
0xfffe76,0xb5d38a,0x086ca5,0x165691,0x1e588e,0x05619b,0x06609a,0x195590,
0x20508e,0x254886,0x2c3678,0x2d3377,0x2e3276,0x2f3075,0x2e3176,0x2e3076,
0x2e2f74,0x2e2f75,0x2d3479,0x2e3076,0x2e2d73,0x313477,0x3f4984,0x374e8a,
0x234b8d,0x175394,0x234f91,0x2b4184,0x303377,0x312e71,0x3c3978,0x3c3876,
0x31306f,0x343372,0x343273,0x342e73,0x342b71,0x33266d,0x382578,0x3e2484,
0x422887,0x4f3f95,0x6561aa,0x8088c2,0xafc2e5,0xd2edfb,0x8bd1f4,0x8ecff4,
0xc3e4f9,0xe8f5fd,0xfeffff,0xffffff,0xffffff,0x000000,0xffffff,0xeef6fe,
0xcae7fb,0x99d3f5,0x7bccf3,0xc3e7fa,0xb4c8e8,0x7b86bf,0x5d59a2,0x4a3a8e,
0x3e2785,0x3b247f,0x332170,0x2d2c72,0x2d3076,0x2f246e,0x231670,0x645c6f,
0xdfda6e,0xede971,0xede876,0xf5ef79,0xf8f27a,0xfaf378,0xfbf478,0xfbf479,
0xfffc77,0xe8f07e,0x448294,0x0c3f8a,0x264c86,0x145795,0x175594,0x224c8b,
0x254889,0x284081,0x2f2c72,0x2f296f,0x30286f,0x2f2970,0x2f2971,0x2f2970,
0x2f2b72,0x2d3076,0x2e2e74,0x2f2c73,0x313375,0x404883,0x364c88,0x224a8b,
0x1a5193,0x254b8e,0x2c3c82,0x2d3075,0x2f2a6e,0x3d3776,0x3b3374,0x30276c,
0x33286e,0x32286d,0x32256c,0x312269,0x33226e,0x3c247e,0x3f2584,0x47358b,
0x58539d,0x7177b5,0x93a6d2,0xcae0f4,0xade0f8,0x7ecaf3,0xb0dcf7,0xd6ecfc,
0xf7fbfe,0xffffff,0xffffff,0x000000,0xffffff,0xfbfdff,0xdff0fc,0xb7def8,
0x81ccf3,0x95d8f6,0xd3e9f9,0x99abd6,0x6b6faf,0x554c99,0x44308a,0x3e2583,
0x39247b,0x2e2b75,0x2c3076,0x31246d,0x28196e,0x2e2471,0xa6a26f,0xf0eb6f,
0xe9e573,0xf0ea76,0xf6ef79,0xf9f279,0xfaf379,0xfaf378,0xfbf479,0xfef678,
0xffff78,0x9eb886,0x14448a,0x204787,0x1b5793,0x1b5393,0x244889,0x264687,
0x293c7f,0x2f286e,0x2f246c,0x2f236b,0x2f236c,0x2f236d,0x2f226b,0x2f246c,
0x2f2970,0x30256e,0x2f256e,0x322c71,0x414280,0x384585,0x264387,0x21498e,
0x284188,0x2f327c,0x2f246f,0x2f1d68,0x3c2a70,0x39256d,0x2e1964,0x301a66,
0x301965,0x301a66,0x392076,0x3f2382,0x422a86,0x4f4494,0x6665a9,0x8391c6,
0xb5cce9,0xc8e9fb,0x7acaf2,0x95d2f5,0xc9e6fa,0xeaf5fe,0xfefeff,0xffffff,
0xffffff,0x000000,0xffffff,0xfffeff,0xf0f7fd,0xcde8fa,0xa1d5f5,0x7acaf2,
0xb7e6fb,0xc0d3ed,0x808cc2,0x625fa5,0x4d3f92,0x412b87,0x3d2682,0x322e7b,
0x2c3075,0x2e246e,0x31206c,0x211570,0x545475,0xd2cf6e,0xede970,0xeae575,
0xf2ec77,0xf6ef79,0xf9f179,0xfaf379,0xfbf478,0xfbf479,0xfbf379,0xfffd78,
0xe8ee7c,0x557586,0x0f3f8a,0x175894,0x175594,0x224a89,0x264787,0x293e7f,
0x2f286f,0x2f266d,0x2e246c,0x2e236b,0x2f226b,0x2e206a,0x2e216b,0x30266e,
0x2f226c,0x2f226b,0x322b70,0x434380,0x394484,0x274085,0x26458b,0x2b3d85,
0x2f317b,0x2f236d,0x301e68,0x402d73,0x3a256d,0x2d1864,0x301c66,0x301b66,
0x301b65,0x351f6f,0x3e2380,0x402783,0x48398d,0x5b56a0,0x757cb8,0x9fb4db,
0xd3e9fa,0x98d8f6,0x7dcaf2,0xb4ddf8,0xdceffc,0xf9fcfe,0xffffff,0xffffff,
0x000000,0xffffff,0xfcfdff,0xe1f1fd,0xbee0f9,0x8bcef3,0x80cff4,0xc6e6f9,
0xa6b8dd,0x7078b5,0x59539d,0x48388e,0x3f2885,0x383085,0x2f327b,0x2e246c,
0x31206b,0x2d206f,0x242479,0x8f8e71,0xebe66c,0xe8e372,0xece775,0xf3ed77,
0xf7ef79,0xf9f279,0xfbf478,0xfbf479,0xfbf379,0xfdf678,0xffff78,0xa7af7d,
0x16438a,0x105193,0x195794,0x224b8a,0x254686,0x293d7e,0x2f296e,0x2f266c,
0x2e246c,0x2f226b,0x2e216a,0x2f206a,0x2f216a,0x30266e,0x2f236c,0x2f226c,
0x332b71,0x444481,0x384282,0x263e84,0x284388,0x2d3b82,0x2f3079,0x2e216b,
0x311d67,0x453176,0x3b256e,0x2c1663,0x301b66,0x301a65,0x321d6a,0x3b217a,
0x3f2383,0x443189,0x534b98,0x696aad,0x899acb,0xc7dcf2,0xbee7fb,0x78c9f2,
0x9fd5f5,0xcbe7fa,0xeff7fe,0xffffff,0xffffff,0x000000,0xffffff,0xf4fafe,
0xd6ecfb,0xafdbf7,0x7ec9f2,0x9adaf8,0xcae2f5,0x94a4d1,0x6a6caf,0x544d9a,
0x44348b,0x3b308a,0x353888,0x312672,0x301e69,0x32246e,0x232279,0x413e75,
0xc3bb6b,0xeae56e,0xe7e272,0xeee875,0xf3ed77,0xf6f079,0xf9f27a,0xfaf479,
0xfbf478,0xfaf378,0xfbf479,0xfbf379,0xfffe78,0xe6e679,0x476386,0x0d408e,
0x1f538f,0x254988,0x284485,0x2a3b7d,0x2f266b,0x2f2269,0x2f216b,0x2f1f69,
0x302069,0x311f68,0x301f69,0x2f236d,0x30226b,0x352c71,0x464480,0x384080,
0x273c82,0x294287,0x2c3b81,0x2f2f77,0x2e206b,0x32206a,0x473578,0x39246d,
0x2c1563,0x301966,0x311b68,0x392078,0x3e2381,0x422c87,0x4f4394,0x6261a6,
0x7d88c0,0xb5c9e7,0xc7eafb,0x81cef4,0x8ecff3,0xbfe1f8,0xe2f1fc,0xfdfeff,
0xffffff,0xffffff,0x000000,0xffffff,0xfefeff,0xedf6fd,0xcee8fb,0xa3d6f7,
0x7ccbf2,0xb1e3fb,0xc5d8ef,0x8896c8,0x6567ab,0x514896,0x40388e,0x383b91,
0x372b7f,0x321e6b,0x31206a,0x312d77,0x211d76,0x63576b,0xd5cf6a,0xe6e16f,
0xe7e373,0xede875,0xf3ed77,0xf6f07a,0xf8f27a,0xf9f279,0xfaf379,0xfaf378,
0xfbf479,0xfcf679,0xffff78,0x9ba77f,0x19408b,0x1b4b8e,0x274b89,0x294787,
0x2c3e7e,0x31296e,0x31266b,0x30256c,0x31226a,0x302169,0x30216a,0x30226a,
0x2f236b,0x30266e,0x2f236c,0x30256c,0x383073,0x484882,0x374180,0x283d82,
0x2a4387,0x2c3c80,0x2e3278,0x2e226c,0x35236d,0x4a387a,0x37236c,0x2c1563,
0x311a68,0x371f74,0x3e2382,0x402985,0x4b3d90,0x5f5ba3,0x767eba,0xa5b8de,
0xceeafa,0x8ed4f5,0x7fcbf3,0xb4ddf8,0xd8edfb,0xf5fafe,0xffffff,0xffffff,
0x000000,0xffffff,0xfdfdff,0xe5f2fd,0xc7e4fb,0x94d1f5,0x79cbf3,0xb9e5fa,
0xbacce9,0x818cc3,0x6262a7,0x4b4897,0x3c3f93,0x3b2f89,0x382378,0x321e6b,
0x2f2b73,0x2c2976,0x25166d,0x8a8069,0xe2dd6b,0xe1dd70,0xe6e272,0xede875,
0xf3ed77,0xf5ef78,0xf7f079,0xf9f279,0xfaf379,0xfaf378,0xfbf278,0xfffc78,
0xf1f17a,0x607785,0x12408d,0x264a89,0x2b4987,0x2c4280,0x322e71,0x322a6e,
0x32286e,0x31266c,0x31256c,0x31236b,0x30266c,0x2f296e,0x30256c,0x30266d,
0x3b3575,0x4b4b84,0x36407f,0x283e81,0x2b4386,0x2d3b81,0x2e3178,0x30226c,
0x331f69,0x38256c,0x311d66,0x301966,0x372074,0x3e2381,0x402785,0x49398e,
0x5a559f,0x7178b5,0x9aabd6,0xcbe3f6,0x9edcf8,0x7ac8f2,0xa8d9f6,0xcfe9fb,
0xeff7fe,0xffffff,0xffffff,0x000000,0xffffff,0xf7fbff,0xdff0fd,0xbce0f9,
0x87cdf3,0x80cef3,0xc2e7fb,0xb3c4e4,0x7b87bf,0x5b5fa6,0x474b9a,0x3e368d,
0x3d2883,0x372177,0x2f2873,0x2f2e73,0x27196e,0x3b2c6c,0xbab368,0xe2dc6a,
0xe0dc6f,0xe7e372,0xeee875,0xf2ec77,0xf5ee78,0xf6f079,0xf8f179,0xf9f279,
0xf9f379,0xfaf379,0xfbf379,0xffff77,0xc5ca7d,0x2e4885,0x1f3581,0x2c3f7f,
0x2c397b,0x31276d,0x30236b,0x31226b,0x31206a,0x301f6a,0x302069,0x301f69,
0x30216a,0x30236b,0x31216a,0x30236b,0x3c3574,0x4c4983,0x343a7c,0x28397e,
0x293e83,0x2d387f,0x2f2d75,0x301f6a,0x2f1c66,0x2d1a63,0x301966,0x371e73,
0x3e2381,0x402785,0x48368c,0x57509c,0x6c71b1,0x92a2d1,0xcadff3,0xaae0f9,
0x78c9f1,0xa0d5f5,0xc9e6fb,0xe7f3fd,0xfdffff,0xffffff,0xffffff,0x000000,
0xffffff,0xf5fafe,0xd7ecfb,0xb3ddf8,0x81cbf3,0x85d1f4,0xc1e5f9,0xacbee1,
0x747fbb,0x585fa6,0x494596,0x402d88,0x3c2582,0x362a7d,0x2f2d76,0x30206b,
0x25136b,0x5c576e,0xccc869,0xe1da6b,0xe0dc70,0xe7e172,0xece774,0xf1eb76,
0xf3ed77,0xf5ef78,0xf6f079,0xf7f079,0xf7f179,0xf8f279,0xf9f279,0xfcf679,
0xffff79,0x7f8b7f,0x1d297c,0x28327a,0x2d3277,0x312169,0x301e68,0x301d67,
0x311d67,0x301e68,0x311d67,0x311f69,0x322069,0x301f69,0x31226a,0x403876,
0x4d4984,0x32367b,0x28387d,0x2a3d82,0x2c387e,0x2f2b73,0x2f1f69,0x2f1c64,
0x311c67,0x371e73,0x3d2280,0x402784,0x47358c,0x564e9b,0x6a6cae,0x8e9dce,
0xc3d8ef,0xb4e4fa,0x79caf2,0x95d1f4,0xc4e4f9,0xe2f1fd,0xfafdfe,0xffffff,
0xffffff,0x000000,0xffffff,0xeff7fd,0xd1eafa,0xafdbf6,0x7ecaf2,0x87d1f5,
0xc1e3f8,0xa6b8dc,0x7180bc,0x595da5,0x4b4092,0x412b87,0x3b2b87,0x353284,
0x322370,0x2f1b68,0x251f74,0x717775,0xd7d267,0xdfd86b,0xdfdb70,0xe6e272,
0xebe574,0xefe875,0xf1eb76,0xf4ee78,0xf5ee78,0xf5ef79,0xf6f079,0xf7f179,
0xf6ef7a,0xfffb79,0xe1e17c,0x4b5079,0x1e2477,0x2e3074,0x31226b,0x31206a,
0x311f68,0x311e68,0x301e67,0x311e68,0x311e67,0x311f68,0x311f69,0x301e68,
0x30226a,0x413a78,0x4b4882,0x313077,0x2a327a,0x2b3a7f,0x2c377d,0x2f276f,
0x2f1d66,0x311c68,0x381f75,0x3d2281,0x402785,0x47368d,0x564e9b,0x696dae,
0x8e9acb,0xc1d6ef,0xb1e2f9,0x77caf2,0x8fd0f4,0xbee1f8,0xdef0fc,0xf7fcff,
0xffffff,0xffffff,0x000000,0xffffff,0xfdffff,0xecf6fe,0xd0e9fa,0xb0dcf7,
0x84ccf3,0x85d1f5,0xbce1f7,0xadbfe2,0x7c88c1,0x5f5fa6,0x4e4294,0x41338b,
0x3a358d,0x382980,0x352273,0x2a2172,0x273882,0x839075,0xd8d165,0xddd66b,
0xddd870,0xe4df70,0xe7e272,0xebe573,0xeee875,0xf0ea75,0xf0eb76,0xf2ec77,
0xf4ee78,0xfffd78,0xc5bf77,0x363474,0x262270,0x301f69,0x301c68,0x301b66,
0x301a66,0x2f1c66,0x2f1d66,0x2f1d67,0x2d1c66,0x302169,0x413d7a,0x494381,
0x302b73,0x2c2c75,0x2c357c,0x2c347a,0x2f256e,0x331e6d,0x392079,0x3e2382,
0x412885,0x4b3b90,0x5a549e,0x6d72b2,0x95a3d0,0xc0d6ef,0xaadff9,0x7bcaf3,
0x92d0f5,0xbde1f9,0xdceffb,0xf6fbff,0xffffff,0xffffff,0x000000,0xffffff,
0xfefeff,0xecf7fd,0xd1eafa,0xb2ddf7,0x85ccf3,0x80cff4,0xbbe4fa,0xbacce9,
0x8490c5,0x6263a8,0x4f4a98,0x424193,0x3d2e89,0x3b2983,0x36287b,0x2a3482,
0x2f4f87,0x899072,0xdbd163,0xdad469,0xdad66e,0xe0db70,0xe3df70,0xe6e172,
0xe9e472,0xeae573,0xece774,0xeee975,0xf0ea76,0xf4ed77,0xfffb78,0x908d77,
0x1d156f,0x2d1a68,0x301b67,0x2f1a67,0x301966,0x301a66,0x2f1a66,0x2e1966,
0x31206a,0x433e7b,0x463f7f,0x2f2871,0x2d2971,0x2c3278,0x2d347c,0x352779,
0x3b217e,0x3f2383,0x432c88,0x4d3f92,0x5e59a2,0x737bb7,0x9dacd5,0xc8dff4,
0xa4ddf7,0x74c7f1,0x96d1f4,0xbee1f9,0xdbeefc,0xf6fbfe,0xffffff,0xffffff,
0x000000,0xffffff,0xfdfeff,0xedf6fd,0xd1e9fa,0xb4def8,0x86ccf4,0x7dcef4,
0xb0e2fa,0xbed4ed,0x909bcb,0x666caf,0x525aa3,0x464093,0x3f348c,0x3b2e87,
0x393b89,0x2e4c8f,0x374f8b,0x959971,0xd8d162,0xd5cf69,0xd6d06c,0xdcd66d,
0xdfda6f,0xe1dd70,0xe4df70,0xe6e172,0xe8e373,0xeae574,0xebe674,0xebe675,
0xf5f075,0xe6e277,0x5c5170,0x210d67,0x301c67,0x2f1a67,0x301966,0x301866,
0x321f69,0x423e79,0x423b7b,0x2d266e,0x2e2972,0x30357e,0x333888,0x3b2a86,
0x3f2785,0x46328a,0x514696,0x605fa5,0x7a84bd,0xaabadd,0xc4e0f5,0x9ddcf8,
0x78c9f1,0x99d3f4,0xbfe2f8,0xdbeefc,0xf4faff,0xffffff,0xffffff,0x000000,
0xffffff,0xfefeff,0xeff7fd,0xd6ebfb,0xb8dff9,0x8ecff4,0x78cbf4,0x9cdcf9,
0xbfdcf3,0xa5b3da,0x7184bf,0x5a5fa7,0x4d4d9a,0x413a90,0x3c3a90,0x3e5695,
0x34539c,0x315194,0x8a9470,0xd3ca5e,0xd0cb65,0xcfc969,0xd5d06b,0xd8d26c,
0xdcd66d,0xdfd96e,0xe0db71,0xe2dd70,0xe3de70,0xe4e071,0xe6e172,0xf7f273,
0xcdc572,0x412f69,0x1f0b66,0x311b67,0x301966,0x2f1765,0x311f68,0x423c78,
0x42387a,0x322875,0x342e81,0x37398d,0x3a388f,0x41318b,0x4b3c90,0x57519c,
0x686cae,0x8b96c8,0xbacce9,0xb9e4f9,0x8bd4f6,0x7ac9f2,0xa0d5f6,0xc5e4f9,
0xdef1fc,0xf6fbff,0xffffff,0xffffff,0x000000,0x4e4e4e,0x3f3f3f,0x272727,
0xfdfdfd,0xffffff,0xfeffff,0xf3f9fd,0xdaeefc,0xc2e2f9,0x9dd4f5,0x7bcaf2,
0x8ed6f7,0xb5e0f8,0xadc8e8,0x899acc,0x6876b3,0x5255a1,0x474998,0x455899,
0x4762a1,0x2a4ea5,0x2a4f96,0x859073,0xc7c15f,0xc7c262,0xc7c366,0xccc768,
0xd0cb68,0xd3ce6a,0xd6d06b,0xd8d26c,0xdad56c,0xddd76d,0xddd96e,0xe1dc6f,
0xf0ec70,0xaaa06e,0x301c67,0x270f65,0x311a66,0x301966,0x301965,0x2f1965,
0x2f1966,0x301865,0x34236e,0x474181,0x4a3c87,0x3a2c84,0x3a348d,0x3e3f93,
0x464294,0x554d9a,0x6361a8,0x7982bc,0xa3b1d7,0xbed8f1,0xa9dffa,0x80d0f4,
0x81cbf2,0xacdaf7,0xc9e6fa,0xe4f2fd,0xf8fcff,0xffffff,0xffffff,0x787878,
0x060606,0x000000,0x131313,0xd1d1d1,0xffffff,0xf7fbff,0xe3f2fc,0xcde8f9,
0xafdbf6,0x8ccff3,0x7dcef4,0x90d6f7,0xb4dbf5,0xaac0e3,0x798bc4,0x5e6aae,
0x5468a5,0x546da7,0x4164a9,0x1f4ea6,0x244d9e,0x6e767a,0xb7b15c,0xc0bc5e,
0xbebb62,0xc2bf64,0xc5c265,0xc8c466,0xcac667,0xcec867,0xd1cb68,0xd2cc69,
0xd4cd6a,0xdbd46b,0xe6e26c,0x9a8f6b,0x2f1966,0x260f65,0x311a66,0x301a67,
0x311a68,0x321c6c,0x361d72,0x381d77,0x3d297f,0x4f468f,0x514492,0x42378e,
0x474597,0x51539e,0x5f62a8,0x777fba,0x9aa7d3,0xbdd1ed,0xb3e0f8,0x8bd5f6,
0x7fccf3,0x99d2f4,0xbbe0f9,0xd5ecfc,0xecf5fd,0xfcfeff,0xffffff,0xffffff,
0x292929,0x101010,0x000000,0x4a4a4a,0x868686,0xffffff,0xfbfdff,0xebf5fd,
0xd5eafb,0xbce0f7,0x9ad4f5,0x79caf2,0x7acff4,0xa3ddf9,0xaacaea,0x8fa3d1,
0x728bba,0x6984b6,0x5278b6,0x3163ac,0x2654a8,0x2a429c,0x676877,0xaba65b,
0xb8b45a,0xb6b35f,0xb9b661,0xbbb861,0xbdbb62,0xc0bc63,0xc3be64,0xc5c166,
0xc7c365,0xc9c367,0xd3cf67,0xd9d668,0x7d6f69,0x291170,0x321974,0x392076,
0x39217a,0x3c217e,0x3e2381,0x3f2484,0x43348a,0x565298,0x5c559d,0x53519d,
0x5f63a9,0x6e78b6,0x8b9aca,0xb2c6e6,0xbee2f7,0x9eddf9,0x79ccf4,0x84ccf3,
0xa8d8f6,0xc8e5fa,0xddeffc,0xf2f8fe,0xfefeff,0xffffff,0xffffff,0xfafafa,
0x2a2a2a,0x000000,0x383838,0xfafafa,0xffffff,0xf3f9fe,0xdef0fc,0xc9e5fa,
0xacdaf8,0x8dd0f4,0x78cbf3,0x79c8f1,0x93c6eb,0xa1c1da,0x96b5d1,0x769fcf,
0x4a7cbc,0x3d6cb3,0x395baa,0x35459d,0x595b80,0x9d9659,0xada852,0xada957,
0xafac5c,0xb0ad5c,0xb3af5e,0xb6b25f,0xb7b45f,0xb9b660,0xbcb862,0xbcb963,
0xc7c662,0xc6c364,0x6f5f76,0x351987,0x3c2185,0x402885,0x432b87,0x45328b,
0x4a3b90,0x534e99,0x6b6eaa,0x7479b3,0x7884bc,0x92a2d0,0xaec3e4,0xb9dbf4,
0x99d9f7,0x7ecef5,0x80ccf4,0x9cd4f5,0xb8def9,0xd3eafb,0xe8f4fd,0xf8fcff,
0xffffff,0xffffff,0xb5b5b5,0x343434,0x171717,0xdadada,0xffffff,0xf9fcff,
0xeaf5fd,0xd6ebfc,0xbde1f8,0xa6d8f6,0x7ec7f0,0x6cbbea,0x78bfe1,0x95cbe1,
0x96cae9,0x72a6d8,0x5c8fca,0x5582bf,0x516db1,0x4e58aa,0x606192,0x878769,
0x9f994c,0xa29c4e,0xa5a056,0xa7a358,0xa8a558,0xaaa659,0xaba85a,0xafaa5b,
0xb0ac5c,0xb2ae5e,0xbebb5d,0xb2ae61,0x695b7f,0x473999,0x4f4499,0x554d99,
0x5f5ba3,0x6c6bad,0x7c83ba,0x96a3cb,0xa4b7da,0xabcae9,0xafd9f5,0x9bd7f6,
0x81cef4,0x7dcbf2,0x98d3f5,0xb3ddf7,0xc9e6fa,0xddf0fc,0xf0f8fd,0xfefeff,
0xffffff,0xffffff,0x5d5d5d,0x959595,0xffffff,0xfeffff,0xf6fbfe,0xe7f4fc,
0xd4ebfb,0xb7ddf7,0x9fceef,0x82bde6,0x76bce2,0x76c6ea,0x60b7e6,0x53aadf,
0x5fa9dc,0x75a0d2,0x879ece,0x8191ce,0x8390c4,0x959894,0x938d58,0x968e42,
0x99934a,0x9b9852,0x9e9a54,0xa09c55,0xa39e56,0xa5a056,0xa6a258,0xb0ac57,
0xa8a25c,0x81818b,0x7c86c6,0x8e9ad0,0x9eadd6,0xaec2e4,0xb3d1ea,0xadd7ec,
0x9cd7f2,0x8ad2f6,0x80cdf3,0x84ccf3,0x9cd4f5,0xb7def8,0xcbe7fa,0xdef0fc,
0xeef7fe,0xfafcfe,0xffffff,0xffffff,0xfdfeff,0xf5fafe,0xdef0fc,0xcde4f8,
0xb4d2ee,0xa1cae5,0x8dc9e9,0x60b6e6,0x38a9e1,0x38aae2,0x5ba8dc,0x80b9e2,
0x97c7ea,0xabcef1,0xb0cdf6,0x9fb8c9,0x969d75,0x8d8643,0x8d863d,0x8f8c4a,
0x908f4f,0x92904f,0x949250,0x979350,0x999551,0x9a9753,0xa29d4f,0xa19e55,
0xa7b4ab,0xb1dbfb,0xa5dfff,0x9cdaf8,0x8fd6f5,0x91d3f1,0x89cff0,0x8fcff4,
0xa9daf7,0xbde0f9,0xcee8fb,0xdff0fc,0xeef8fd,0xfafdff,0xffffff,0xffffff,
0xf7fbfe,0xeef6fe,0xdae6f3,0xc2d8e6,0xb5d8ea,0x94c9ec,0x6bb6e6,0x58ace1,
0x58a4d9,0x68aee0,0x76bfec,0x8accf0,0x98d6f6,0x8ad5ff,0x8accee,0x7daaac,
0x78875e,0x7f7936,0x837e3d,0x868548,0x89884c,0x8a894b,0x8c8a4b,0x8e8c4d,
0x8f8e4e,0x949048,0x99964e,0x89ad9e,0x82cff9,0x89d2fb,0x98d2f6,0xb4dff5,
0xbde1f5,0xc8e7fa,0xd8edfc,0xe5f2fc,0xf0f8fe,0xfafdff,0xffffff,0xffffff,
0xfafcfe,0xf1f2f7,0xdbe3e9,0xd1e4eb,0xb8d8f1,0x91c4eb,0x80b4e1,0x80aada,
0x97bbe2,0xa5cff0,0xa3d3f3,0xa0d4f3,0x8ecef3,0x8ad2fc,0x89d5ff,0x86c2de,
0x859f93,0x7c7f51,0x777235,0x7b783c,0x807e46,0x838248,0x848348,0x858449,
0x87864a,0x878341,0x85813e,0x97ab9b,0xc0e1f5,0xd8f2ff,0xdceffa,0xe1f0fa,
0xedf8fd,0xf7fbfe,0xfdfeff,0xffffff,0xffffff,0xe2e2e2,0x7f7f7f,0xe2e2e2,
0xffffff,0xfbfdfe,0xf9fcfe,0xf5f5f9,0xe4e7ed,0xdfe9ed,0xcfe3f4,0xaecdec,
0x9bbde2,0x9ab1da,0xb5c5e4,0xd1e4f6,0xd4e9fa,0xd2e8f9,0xc7e5fa,0xc0e1f8,
0xbce2fa,0xb9e5ff,0xb5e1fe,0xabcad6,0x94a08b,0x797849,0x706c33,0x736f36,
0x7b7840,0x7e7c45,0x7e7d46,0x807e47,0x817d40,0x807c3d,0x9b9f7f,0xe2e8e5,
0xffffff,0xfcfeff,0xffffff,0xffffff,0xc4c4c4,0x000000,0xc4c4c4,0xffffff,
0xfdfeff,0xf8fcff,0xf7f8fc,0xeaebf1,0xe5ebef,0xd8e9f5,0xbad1ec,0xabc3e3,
0xadbbdd,0xc9cee6,0xf0f0f9,0xf6f9fe,0xf4f9fe,0xf0f8fe,0xecf6fe,0xeaf5fc,
0xe7f3fc,0xe5f5ff,0xe5f9ff,0xe1f2f9,0xd1dbd7,0xb5b9a6,0x8c8b65,0x716d3a,
0x6a6732,0x706d38,0x77743f,0x787643,0x77743b,0x736f33,0x8d8960,0xcac8b8,
0xfdfcfd,0xffffff,0xffffff,0xc4c4c4,0x000000,0xc4c4c4,0xffffff,0xfdfeff,
0xf9fcff,0xfafbff,0xf2f3f6,0xeaeff0,0xdeedf5,0xc1d6ee,0xb0c5e5,0xb3c1e2,
0xcdd0e7,0xf5f2f8,0xfefefe,0xffffff,0xfeffff,0xfefefe,0xfeffff,0xffffff,
0xfcfeff,0xf1f1ed,0xccccbd,0xa29f83,0x7a794d,0x676531,0x686631,0x6d6b38,
0x6f6c39,0x68662e,0x6f6d3a,0xa19f86,0xe8e8e6,0xffffff,0xffffff,0xc4c4c4,
0x000000,0xc4c4c4,0xffffff,0xfcfdff,0xf7f8f9,0xecf2f1,0xe2eef4,0xcbdef2,
0xb5cce8,0xb6c6e5,0xced3e9,0xf3f2f9,0xfcfdfe,0xffffff,0xfbfbfa,0xe4e4dd,
0xbcbca9,0x9a997c,0x78774c,0x6c6a3a,0x646431,0x605e29,0x5b5923,0x787751,
0xbdbdaf,0xf2f1f0,0xfefefe,0xffffff,0xffffff,0xc4c4c4,0x000000,0xc4c4c4,
0xffffff,0xfeffff,0xfafcfb,0xeff5f3,0xe8f2f5,0xd6e8f8,0xc0d6f0,0xbccbe7,
0xcfd5eb,0xf3f3f9,0xfcfcfe,0xffffff,0xfefefe,0xeae9e4,0xcecec0,0xacac95,
0x939274,0x81815c,0x6a683c,0x5f5e30,0x7a7857,0xb2b1a1,0xdeded8,0xf5f5f5,
0xfcfcfc,0xfefefe,0xffffff,0xffffff,0xc4c4c4,0x000000,0xc4c4c4,0xffffff,
0xfefefe,0xf1f7f5,0xecf4f4,0xe1edfa,0xcbdef3,0xc4d5ee,0xd0daee,0xf1f4fa,
0xfcfdfe,0xffffff,0xfdfdfe,0xeeeeeb,0xd9d9d0,0xbbbaa8,0x9f9e83,0x919073,
0xa5a391,0xd3d2ce,0xededea,0xf7f6f5,0xffffff,0xffffff,0xa2a2a2,0x7f7f7f,
0x626262,0x000000,0x626262,0x7f7f7f,0xa2a2a2,0xffffff,0xf7fbf9,0xf0f6f6,
0xe8f1f8,0xd8e3f3,0xcee0f5,0xd6e4f5,0xf2f3fa,0xfdfdfe,0xffffff,0xfffffe,
0xf5f5f2,0xf2f2ef,0xf5f5f2,0xf8f7f5,0xfefefe,0xffffff,0xffffff,0x454545,
0x000000,0x454545,0xffffff,0xf9fcfa,0xeff4f4,0xeff5f9,0xe4ebf6,0xd9e7f8,
0xddecfa,0xf3f6fa,0xfefdfe,0xfffffe,0xffffff,0xffffff,0xb7b7b7,0xa2a2a2,
0xb7b7b7,0xffffff,0xfafcfb,0xeef4f2,0xf3f6f8,0xf1f6fc,0xe5f1fb,0xe6f1fa,
0xf4f8fc,0xfefeff,0xfffffe,0xffffff,0xffffff,0xfefefe,0xf5f9f7,0xf0f4f4,
0xf7f9fd,0xf4f8fd,0xf0f2f9,0xf6f8fc,0xfefeff,0xffffff,0xffffff,0xfcfefc,
0xf5f8f6,0xf9fbfb,0xfcfcfe,0xfafafc,0xfbfbfd,0xfefefe,0xffffff,0xffffff,
0xbfbfbf,0x7f7f7f,0x919191,0xffffff,0xfefefe,0xfcfdfc,0xfefffe,0xffffff,
0xfdfdfd,0xffffff,0xbfbfbf,0x7f7f7f,0xffffff,0xa6a6a6,0x7f7f7f,0x848484,
0xfbfbfb,0xffffff,0xffffff,0x7f7f7f,0x000000,0x222222,0xffffff,0x787878,
0x000000,0xffffff,0xfeffff,0xffffff,0x272727,0xd1d1d1,0xffffff,0x7f7f7f,
0x000000,0xffffff,0xe2e2e2,0x080808,0x000000,0x6a6a6a,0xffffff,0xffffff,
0x7f7f7f,0x000000,0xa2a2a2,0xaeaeae,0xffffff,0xfafafa,0x292929,0x060606,
0x000000,0xffffff,0x3f3f3f,0x131313,0x868686,0xfafafa,0xffffff,0x7f7f7f,
0x000000,0xffffff,0x5d5d5d,0x000000,0x111111,0xeaeaea,0xffffff,0xffffff,
0x7f7f7f,0x000000,0xffffff,0xb5b5b5,0x2a2a2a,0x101010,0x000000,0xffffff,
0x4e4e4e,0x000000,0x4a4a4a,0x383838,0xdadada,0xffffff,0x7f7f7f,0x000000,
0x858585,0x000000,0x0c0c0c,0xafafaf,0xffffff,0xffffff,0x7f7f7f,0x000000,
0xffffff,0x5d5d5d,0x000000,0x171717,0x959595,0xffffff,0x7f7f7f,0x000000,
0x494949,0xf7f7f7,0xffffff,0xffffff,0x7f7f7f,0x000000,0xffffff,0x959595,
0x000000,0x343434,0x5d5d5d,0xffffff,0x7f7f7f,0x000000,0xdddddd,0xbfbfbf,
0x0c0c0c,0x000000,0x9d9d9d,0xffffff,0xffffff,0x7f7f7f,0x000000,0xffffff,
0xdadada,0x383838,0x4a4a4a,0x000000,0x4e4e4e,0xffffff,0x000000,0x101010,
0x2a2a2a,0xb5b5b5,0xffffff,0x7f7f7f,0x000000,0xc4c4c4,0xb3b3b3,0x080808,
0x000000,0x888888,0xffffff,0xffffff,0x7f7f7f,0x000000,0xffffff,0xfafafa,
0x868686,0x131313,0x3f3f3f,0xffffff,0x000000,0x060606,0x292929,0xfafafa,
0xffffff,0x7f7f7f,0x000000,0x111111,0xe2e2e2,0xffffff,0xffffff,0xd0d0d0,
0xa2a2a2,0xffffff,0xd1d1d1,0x272727,0xffffff,0x000000,0x787878,0xffffff,
0xd0d0d0,0xa2a2a2,0xaeaeae,0xeaeaea,0xffffff,0xffffff,0xfdfdfd,0xffffff,
0xffffff,0xffffff,0xffffff,};